import pytest

from gaphor.core.styling import PrefersColorScheme, compile_style_sheet
from gaphor.core.styling.selectors import SelectorError


class Node:
    def __init__(
        self,
        name,
        parent=None,
        children=None,
        classes=None,
        attributes=None,
        state=(),
        pseudo=None,
    ):
        if attributes is None:
            attributes = {}
        self._name = name
        self._parent = parent
        self._children = children or []
        self._classes = classes or []
        self._attributes = attributes
        self._state = state
        self.pseudo = pseudo

        if parent:
            parent._children.append(self)  # noqa: SLF001
        for c in self._children:
            c._parent = self  # noqa: SLF001

    def name(self):
        return self._name

    def parent(self):
        return self._parent

    def children(self):
        return iter(self._children)

    def classes(self):
        return self._classes

    def attribute(self, name):
        return self._attributes.get(name, "") if name in self._attributes else None

    def state(self):
        return self._state


def test_node_test_object_parent_child():
    c = Node("child")
    p = Node("parent", children=[c])

    assert c.name() == "child"
    assert p.name() == "parent"
    assert c.parent() is p
    assert c in p.children()


def test_node_test_object_child_parent():
    p = Node("parent")
    c = Node("child", parent=p)

    assert c.name() == "child"
    assert p.name() == "parent"
    assert c.parent() is p
    assert c in p.children()


def test_select_all():
    css = "* {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector("any")


def test_select_name():
    css = "classitem {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem"))
    assert not selector(Node("packageitem"))


def test_select_inside_combinator():
    css = "classitem nested {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("nested", parent=Node("classitem")))
    assert selector(
        Node(
            "nested",
            parent=Node("other", parent=Node("classitem")),
        )
    )
    assert not selector(Node("nested"))
    assert not selector(Node("classitem"))


def test_select_parent_combinator():
    css = "classitem > nested {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("nested", parent=Node("classitem")))
    assert not selector(
        Node(
            "nested",
            parent=Node("other", parent=Node("classitem")),
        )
    )
    assert not selector(Node("nested"))
    assert not selector(Node("classitem"))


def test_select_sibling_combinator():
    css = "previous + sibling {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("sibling", parent=Node("parent", children=[Node("previous")])))
    assert not selector(
        Node("sibling", parent=Node("parent", children=[Node("other")]))
    )
    assert not selector(
        Node("other", parent=Node("parent", children=[Node("previous")]))
    )
    assert not selector(Node("sibling"))


def test_class():
    css = ".element {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("item", classes=["element"]))


def test_attributes():
    css = "classitem[subject] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "val"}))
    assert not selector(Node("classitem"))


def test_attribute_equal():
    css = "classitem[subject=foo] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "foo"}))
    assert not selector(Node("classitem", attributes={"subject": "bar"}))
    assert not selector(Node("classitem"))


def test_attribute_in_list():
    css = "classitem[subject~=foo] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "foo"}))
    assert selector(Node("classitem", attributes={"subject": "foo bar"}))
    assert not selector(Node("classitem", attributes={"subject": "bar"}))
    assert not selector(Node("classitem"))


def test_attribute_starts_with():
    css = "classitem[subject^=foo] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "foo"}))
    assert selector(Node("classitem", attributes={"subject": "foomania"}))
    assert not selector(Node("classitem", attributes={"subject": "not foo"}))
    assert not selector(Node("classitem"))


def test_attribute_starts_with_dash():
    css = "classitem[subject|=foo] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "foo"}))
    assert selector(Node("classitem", attributes={"subject": "foo-mania"}))
    assert not selector(Node("classitem", attributes={"subject": "foomania"}))
    assert not selector(Node("classitem"))


def test_attribute_ends_with():
    css = "classitem[subject$=foo] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "foo"}))
    assert selector(Node("classitem", attributes={"subject": "manicfoo"}))
    assert not selector(Node("classitem", attributes={"subject": "fooless"}))
    assert not selector(Node("classitem"))


def test_attribute_contains():
    css = "classitem[subject*=foo] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject": "foo"}))
    assert selector(Node("classitem", attributes={"subject": "be foo-ish"}))
    assert not selector(Node("classitem", attributes={"subject": "fobic"}))
    assert not selector(Node("classitem"))


def test_attributes_with_dots():
    css = "classitem[subject.members] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"subject.members": "foo"}))


def test_attributes_are_lower_case():
    css = "ClassItem[MixedCase] {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", attributes={"mixedcase": "foo"}))


def test_empty_pseudo_selector():
    css = ":empty {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("node"))
    assert not selector(Node("node", children=[Node("child")]))


def test_empty_pseudo_selector_with_name():
    css = "node:empty {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("node"))
    assert not selector(Node("node", children=[Node("child")]))


def test_root_pseudo_selector_with_name():
    css = ":root {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("node"))
    assert not selector(Node("node", parent=Node("child")))


def test_first_child_selector():
    css = "node:first-child {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert not selector(Node("node", parent=Node("parent", children=[Node("node")])))
    assert selector(Node("node", parent=Node("parent")))
    assert not selector(Node("sibling"))


@pytest.mark.parametrize(
    "state",
    ["hover", "focus", "active", "drop", "disabled"],
)
def test_hovered_pseudo_selector(state):
    css = f":{state} {{}}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("node", state=(state,)))
    assert selector(Node("node", state=(state, "other-state")))
    assert not selector(Node("node", state=()))


def test_is_pseudo_selector():
    css = "classitem:is(:hover, :active) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", state=("hover",)))
    assert selector(Node("classitem", state=("active",)))
    assert selector(Node("classitem", state=("hover", "active")))
    assert not selector(Node("classitem"))


def test_not_pseudo_selector():
    css = "classitem:not(:hover) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem"))
    assert selector(Node("classitem", state=("active")))
    assert not selector(Node("classitem", state=("hover",)))
    assert not selector(Node("classitem", state=("hover", "active")))


def test_has_pseudo_selector():
    css = "classitem:has(nested) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", children=[Node("nested")]))
    assert selector(
        Node("classitem", children=[Node("middle", children=[Node("nested")])])
    )
    assert not selector(Node("classitem"))


def test_has_pseudo_selector_with_wildcard():
    css = "classitem:has(*) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(Node("classitem", children=[Node("nested")]))
    assert selector(
        Node("classitem", children=[Node("middle", children=[Node("nested")])])
    )
    assert not selector(Node("classitem"))


def test_has_pseudo_selector_with_nested_selector():
    css = "classitem:has(middle nested) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(
        Node("classitem", children=[Node("middle", children=[Node("nested")])])
    )
    assert not selector(Node("classitem", children=[Node("nested")]))
    assert selector(
        Node(
            "classitem",
            children=[
                Node("middle", children=[Node("other", children=[Node("nested")])])
            ],
        )
    )


def test_has_pseudo_selector_with_complex_selector():
    css = "classitem:has(middle > nested) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(
        Node("classitem", children=[Node("middle", children=[Node("nested")])])
    )
    assert not selector(Node("classitem", children=[Node("nested")]))
    assert not selector(
        Node(
            "classitem",
            children=[
                Node("middle", children=[Node("other", children=[Node("nested")])])
            ],
        )
    )


# TODO: customize parser to allow expressions like ":has(> nested)"
def test_has_pseudo_selector_with_combinator_is_not_supported():
    # NB. This is according to the CSS spec, but our parser is not
    # able to deal with this. This test is just here to illustrate.
    css = "classitem:has(> nested) {}"

    error, _payload = next(compile_style_sheet(css))

    assert error == "error"


def test_parse_empty_pseudo_element():
    css = "node::after {}"

    selector, declarations = next(compile_style_sheet(css))

    assert declarations == {}
    assert not selector(Node("node"))
    assert selector(Node("node", pseudo="after"))


def test_has_and_is_selector():
    css = "node:has(:is(:hover)) {}"

    selector, _declarations = next(compile_style_sheet(css))

    assert selector(
        Node(
            "node",
            children=[Node("foo", children=[Node("bar", state=("hover",))])],
        )
    )


@pytest.mark.parametrize(
    "css",
    [
        "@media { node { color: blue; } }",
        "@media(prefers-color-scheme = dark) { node { color: blue; } }",
        "@media prefers-color-scheme = dark { node { color: blue; } }",
        "@media(dark-mode) { node { color: blue; } }",
        "@media dark-mode { node { color: blue; } }",
    ],
)
def test_media_query(css):
    selector, _declarations = next(
        compile_style_sheet(css, prefers_color_scheme=PrefersColorScheme.DARK)
    )

    assert selector(Node("node"))


@pytest.mark.parametrize(
    "css,exc_type",
    [
        ["@media(prefers-color-scheme { * { color: blue; } }", StopIteration],
        ["@media prefers-color-scheme true) { * { color: blue; } }", SelectorError],
        ["@media((other-query = = dark) { * { color: blue; } }", StopIteration],
        ["@media{(prefers-color-scheme = dark) { color }", StopIteration],
    ],
)
def test_invalid_media_query(css, exc_type):
    with pytest.raises(exc_type):
        next(compile_style_sheet(css))



import pytest
from gi.repository import Adw

from gaphor.event import Notification
from gaphor.ui.notification import InAppNotifier


@pytest.fixture
def in_app_notifier():
    return InAppNotifier(Adw.ToastOverlay.new())


def test_notifier_can_handle_message(in_app_notifier):
    toast = in_app_notifier.handle(Notification("a message"))

    assert toast


def test_notifier_can_collapse_multiple_messages(in_app_notifier):
    toast1 = in_app_notifier.handle(Notification("a message"))
    toast2 = in_app_notifier.handle(Notification("a message"))

    assert toast1
    assert not toast2


def test_notifier_show_messages_when_previous_message_is_dismissed(in_app_notifier):
    toast1 = in_app_notifier.handle(Notification("a message"))
    toast1.dismiss()
    toast2 = in_app_notifier.handle(Notification("a message"))

    assert toast2



# ruff: noqa: T201

from __future__ import annotations

import os
import shutil
import subprocess
from pathlib import Path

version = subprocess.run(
    ["poetry", "version", "-s"],
    encoding="utf-8",
    capture_output=True,
    text=True,
    check=True,
).stdout.rstrip()


def clean_files(paths: list[Path]) -> None:
    print("Cleaning files")
    for path in paths:
        if path.is_dir():
            shutil.rmtree(path, ignore_errors=True)
        elif path.is_file():
            path.unlink(missing_ok=True)


def build_installer(
    icon: Path, nsi: Path, files_to_package: Path, output_file: Path
) -> None:
    print("Building Installer")
    shutil.copy(icon, files_to_package / "gaphor.ico")
    os.chdir(files_to_package)
    subprocess.run(
        [
            "makensis",
            "-NOCD",
            f"-DVERSION={version}",
            str(nsi),
        ],
        encoding="utf-8",
        capture_output=True,
        text=True,
        check=True,
    )

    shutil.move(str(files_to_package / "gaphor-LATEST.exe"), output_file)


def concatenate_files(input_files: list[Path], output_file: Path) -> None:
    print(f"Opening {output_file} for write")
    with output_file.open("wb") as writer:
        for input_file in input_files:
            print(f"Writing {input_file}")
            with input_file.open("rb") as reader:
                shutil.copyfileobj(reader, writer)


def unix2dos(file_path: Path) -> None:
    win_line_ending = b"\r\n"
    unix_line_ending = b"\n"

    with file_path.open("rb") as open_file:
        content = open_file.read()
    content = content.replace(win_line_ending, unix_line_ending)
    with file_path.open("wb") as open_file:
        open_file.write(content)


def build_portable_installer(
    symlink_path: Path,
    readme_path: Path,
    files_to_package: Path,
    portable_path: Path,
    payload_path: Path,
    seven_zip_path: Path,
    output_file: Path,
) -> None:
    print("Building portable installer")
    shutil.copy(symlink_path, portable_path / symlink_path.name)
    shutil.copy(readme_path, portable_path / "README.txt")
    unix2dos(portable_path / "README.txt")
    config_path = portable_path / "config"
    config_path.mkdir()
    shutil.copytree(files_to_package, portable_path / "data")

    subprocess.run(
        [str(seven_zip_path), "a", str(payload_path), str(portable_path)], check=True
    )
    if payload_path.is_file():
        print(f"Payload 7z archive found at {payload_path}")
    else:
        print("Payload 7z archive not found")
    sfx_path = seven_zip_path.parent / "7z.sfx"
    if sfx_path.is_file():
        print(f"Sfx file found at {sfx_path}")
    else:
        print("Sfx file not found")
    concatenate_files([seven_zip_path.parent / "7z.sfx", payload_path], output_file)


def main():
    working_dir: Path = Path.cwd()
    dist: Path = working_dir / "dist"
    Path(dist / "gaphor").mkdir(parents=True, exist_ok=True)

    portable: Path = dist / f"gaphor-{version}-portable"
    payload: Path = dist / "payload.7z"

    clean_files([portable, payload])

    icon = working_dir / "_packaging" / "windows" / "gaphor.ico"
    nsi = working_dir / "_packaging" / "windows" / "win_installer.nsi"
    gaphor_files: Path = dist / "gaphor"
    installer = dist / f"gaphor-{version}-installer.exe"
    build_installer(icon, nsi, gaphor_files, installer)

    portable.mkdir(parents=True)
    symlink = working_dir / "_packaging" / "windows" / "gaphor.lnk"
    readme = working_dir / "_packaging" / "windows" / "README-PORTABLE.txt"
    seven_zip: Path = Path("C:/Program Files/7-Zip/7z.exe")
    installer = dist / f"gaphor-{version}-portable.exe"
    build_portable_installer(
        symlink, readme, gaphor_files, portable, payload, seven_zip, installer
    )

    clean_files([portable, payload])
    print("Windows Installer builds are complete!")


if __name__ == "__main__":
    main()



# ruff: noqa: F401,F811

import pytest

from gaphor import UML
from gaphor.UML.actions.actionseditors import (
    activity_parameter_node_item_editor,
    fork_node_item_editor,
)


@pytest.mark.asyncio
async def test_activity_parameter_node_item_editor(
    diagram, element_factory, view, event_manager
):
    item = diagram.create(
        UML.actions.ActivityParameterNodeItem,
        subject=element_factory.create(UML.ActivityParameterNode),
    )
    item.subject.parameter = element_factory.create(UML.Parameter)
    view.selection.hovered_item = item
    await view.update()
    result = activity_parameter_node_item_editor(item, view, event_manager)

    assert result is True


def test_fork_node_item_editor(diagram, element_factory, view, event_manager):
    item = diagram.create(
        UML.actions.ForkNodeItem, subject=element_factory.create(UML.ForkNode)
    )
    view.selection.hovered_item = item
    result = fork_node_item_editor(item, view, event_manager)

    assert result is False


@pytest.mark.asyncio
async def test_join_node_item_editor(diagram, element_factory, view, event_manager):
    item = diagram.create(
        UML.actions.ForkNodeItem, subject=element_factory.create(UML.JoinNode)
    )
    view.selection.hovered_item = item
    await view.update()
    result = fork_node_item_editor(item, view, event_manager)

    assert result is True



from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.UML.deployments.connector import ConnectorItem


def test_create(create):
    """Test creation of connector item."""
    conn = create(ConnectorItem, UML.Connector)
    assert conn.subject is not None


def test_persistence(create, element_factory, saver, loader):
    """Test connector item saving/loading."""
    conn = create(ConnectorItem, UML.Connector)

    end = element_factory.create(UML.ConnectorEnd)
    conn.subject.end = end

    data = saver()
    assert end.id in data

    loader(data)
    diagram = next(element_factory.select(Diagram))
    assert diagram.select(ConnectorItem)
    assert element_factory.lselect(UML.ConnectorEnd)



"""Test include item connections."""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect, get_connected
from gaphor.UML.usecases.include import IncludeItem
from gaphor.UML.usecases.usecase import UseCaseItem


def test_use_case_glue(create):
    """Test "include" gluing to use cases."""

    uc1 = create(UseCaseItem, UML.UseCase)
    include = create(IncludeItem)

    glued = allow(include, include.head, uc1)
    assert glued


def test_use_case_connect(create):
    """Test connecting "include" to use cases."""
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)
    include = create(IncludeItem)

    connect(include, include.head, uc1)
    assert get_connected(include, include.head), uc1

    connect(include, include.tail, uc2)
    assert get_connected(include, include.tail), uc2


def test_use_case_reconnect(create):
    """Test reconnecting use cases with "include"."""
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)
    uc3 = create(UseCaseItem, UML.UseCase)
    include = create(IncludeItem)

    # connect: uc1 -> uc2
    connect(include, include.head, uc1)
    connect(include, include.tail, uc2)
    e = include.subject

    # reconnect: uc1 -> uc2
    connect(include, include.tail, uc3)

    assert e is not include.subject
    assert include.subject.addition is uc1.subject
    assert include.subject.includingCase is uc3.subject


def test_use_case_disconnect(create):
    """Test disconnecting "include" from use cases."""
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)
    include = create(IncludeItem)

    connect(include, include.head, uc1)
    connect(include, include.tail, uc2)

    disconnect(include, include.head)
    assert get_connected(include, include.head) is None
    assert include.subject is None

    disconnect(include, include.tail)
    assert get_connected(include, include.tail) is None



from gaphor import UML


def test_qualified_name():
    p1 = UML.Package()
    p2 = UML.Package()
    p3 = UML.Package()

    p1.name = "package1"
    p2.name = "package2"
    p3.name = "package3"
    p3.nestingPackage = p2
    p2.nestingPackage = p1

    assert p3.qualifiedName == ["package1", "package2", "package3"]


def test_partly_connected_component_required_relation():
    usage = UML.Usage()
    component = UML.Component()

    component.clientDependency = usage

    assert not component.required


def test_component_required_relation():
    usage = UML.Usage()
    component = UML.Component()
    interface = UML.Interface()

    component.clientDependency = usage
    usage.supplier = interface

    assert [interface] in component.required


def test_component_provided_relation():
    realization = UML.InterfaceRealization()
    component = UML.Component()
    interface = UML.Interface()

    component.interfaceRealization = realization
    realization.contract = interface

    assert interface in component.provided



from gaphor.plugins.console.console import docstring_dedent


def test_docstring_with_leading_space():
    docstr = """\
        line one
        line two
        """

    expected = "line one\nline two\n"

    assert docstring_dedent(docstr) == expected


def test_docstring_without_leading_space():
    docstr = """line one
        line two
        """

    expected = "line one\nline two\n"

    assert docstring_dedent(docstr) == expected


def test_docstring_without_leading_space_with_blank_line():
    docstr = """line one

        line two
        """

    expected = "line one\n\nline two\n"

    assert docstring_dedent(docstr) == expected



class raises_exception_group:
    def __init__(
        self,
        expected_exception,
    ) -> None:
        self.expected_exception = expected_exception

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type,
        exc_val,
        exc_tb,
    ) -> bool:
        return (
            any(isinstance(exc, self.expected_exception) for exc in exc_val.exceptions)
            if issubclass(exc_type, ExceptionGroup)
            else False
        )



from gaphor import UML
from gaphor.diagram.group import group, ungroup
from gaphor.UML.classes import ComponentItem
from gaphor.UML.usecases import UseCaseItem


def test_grouping(create):
    """Test adding a use case to a subsystem."""
    s = create(ComponentItem, UML.Component)
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)

    group(s.subject, uc1.subject)
    assert len(uc1.subject.subject) == 1
    group(s.subject, uc2.subject)
    assert len(uc2.subject.subject) == 1

    assert len(s.subject.useCase) == 2


def test_grouping_with_namespace(create):
    """Test adding a use case to a subsystem (with namespace)"""
    s = create(ComponentItem, UML.Component)
    uc = create(UseCaseItem, UML.UseCase)

    group(s.subject, uc.subject)
    assert len(uc.subject.subject) == 1
    assert s.subject in uc.subject.subject


def test_ungrouping(create):
    """Test removal of use case from subsystem."""
    s = create(ComponentItem, UML.Component)
    uc1 = create(UseCaseItem, UML.UseCase)
    uc2 = create(UseCaseItem, UML.UseCase)

    assert group(s.subject, uc1.subject)
    assert group(s.subject, uc2.subject)

    ungroup(s.subject, uc1.subject)
    assert len(uc1.subject.subject) == 0
    assert len(s.subject.useCase) == 1

    ungroup(s.subject, uc2.subject)
    assert len(uc2.subject.subject) == 0
    assert len(s.subject.useCase) == 0



from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.general import CommentLineItem
from gaphor.UML.usecases.actor import ActorItem


def test_create_actor(create, diagram):
    actor_item = create(ActorItem, UML.Actor)
    actor_item.subject.name = "Actor"

    diagram.update({actor_item})


def test_connect_to_lower_port(create, diagram):
    actor_item = create(ActorItem, UML.Actor)
    actor_item.subject.name = "Actor Name"
    line_item = create(CommentLineItem)
    line_head, line_tail = line_item.handles()
    line_head.pos = (50, 100)
    line_tail.pos = (50, 100)

    diagram.update({actor_item, line_item})

    connect(line_item, line_head, actor_item)

    cinfo = diagram.connections.get_connection(line_head)
    assert cinfo.port is actor_item.ports()[-1]


def test_save_and_load(create, diagram, element_factory, saver, loader):
    actor_item = create(ActorItem, UML.Actor)
    actor_item.subject.name = "Actor Name"
    line_item = create(CommentLineItem)
    line_head, line_tail = line_item.handles()
    line_head.pos = (50, 100)
    line_tail.pos = (50, 100)

    diagram.update({actor_item, line_item})

    connect(line_item, line_head, actor_item)

    diagram.update({actor_item, line_item})

    data = saver()
    loader(data)

    new_diagram = next(element_factory.select(Diagram))
    new_actor_item = next(element_factory.select(ActorItem))
    new_line_item = next(element_factory.select(CommentLineItem))
    cinfo = new_diagram.connections.get_connection(new_line_item.head)

    assert cinfo.port is new_actor_item.ports()[-1]



import pytest

from gaphor import UML
from gaphor.codegen.coder import (
    associations,
    bases,
    class_declaration,
    is_in_profile,
    is_in_toplevel_package,
    is_simple_type,
    load_model,
    load_modeling_language,
    order_classes,
    resolve_attribute_type_values,
    superset_attribute,
    variables,
)
from gaphor.core.format import parse
from gaphor.core.modeling import ElementFactory
from gaphor.core.modeling.modelinglanguage import (
    CoreModelingLanguage,
    MockModelingLanguage,
)
from gaphor.diagram.general.modelinglanguage import GeneralModelingLanguage
from gaphor.UML.modelinglanguage import UMLModelingLanguage


@pytest.fixture(scope="session")
def core_metamodel():
    return load_model(
        "models/Core.gaphor",
        MockModelingLanguage(
            CoreModelingLanguage(), GeneralModelingLanguage(), UMLModelingLanguage()
        ),
    )


@pytest.fixture(scope="session")
def uml_metamodel():
    return load_model(
        "models/UML.gaphor",
        MockModelingLanguage(
            CoreModelingLanguage(), GeneralModelingLanguage(), UMLModelingLanguage()
        ),
    )


def test_load_modeling_language():
    ml = load_modeling_language("Core")

    assert ml.__class__.__name__ == "CoreModelingLanguage"


def test_coder_write_class():
    class_ = UML.Class()
    class_.name = "TestClass"

    class_def = class_declaration(class_)

    assert class_def == "class TestClass():"


def test_coder_write_class_no_attributes():
    class_ = UML.Class()

    attr_def = list(variables(class_))

    assert not attr_def


def create_attribute(s: str, element_factory=None):
    if element_factory:
        attr = element_factory.create(UML.Property)
    else:
        attr = UML.Property()
    parse(attr, s)
    return attr


def create_literal(s: str, element_factory):
    literal = UML.EnumerationLiteral()
    parse(literal, s)
    return literal


def test_coder_write_class_with_attributes():
    class_ = UML.Class()
    class_.ownedAttribute = create_attribute("first: str")
    class_.ownedAttribute = create_attribute("second: int")

    attr_def = list(variables(class_))

    assert attr_def == [
        'first: _attribute[str] = _attribute("first", str)',
        'second: _attribute[int] = _attribute("second", int)',
    ]


def test_coder_write_class_with_enumeration(element_factory: ElementFactory):
    class_ = element_factory.create(UML.Class)
    class_.ownedAttribute = create_attribute("first: EnumKind", element_factory)

    enum = element_factory.create(UML.Enumeration)
    enum.name = "EnumKind"
    enum.ownedLiteral = create_literal("in", element_factory)
    enum.ownedLiteral = create_literal("out", element_factory)

    resolve_attribute_type_values(element_factory)

    attr_def = list(variables(class_))

    assert attr_def == ['first = _enumeration("first", EnumKind, EnumKind.in_)']


def test_coder_write_class_with_enumeration_and_default_value(
    element_factory: ElementFactory,
):
    class_ = element_factory.create(UML.Class)
    class_.ownedAttribute = create_attribute("first: EnumKind = out", element_factory)

    enum = element_factory.create(UML.Enumeration)
    enum.name = "EnumKind"
    enum.ownedLiteral = create_literal("in", element_factory)
    enum.ownedLiteral = create_literal("out", element_factory)

    resolve_attribute_type_values(element_factory)

    attr_def = list(variables(class_))

    assert attr_def == ['first = _enumeration("first", EnumKind, EnumKind.out)']


@pytest.fixture
def navigable_association(element_factory):
    class_a = element_factory.create(UML.Class)
    class_a.name = "A"
    class_b = element_factory.create(UML.Class)
    class_b.name = "B"
    association = UML.recipes.create_association(class_a, class_b)
    UML.recipes.set_navigability(association, association.memberEnd[0], True)
    UML.recipes.set_navigability(association, association.memberEnd[1], True)
    association.memberEnd[0].name = "a"
    association.memberEnd[1].name = "b"
    return association


def test_coder_write_class_with_n_m_association(navigable_association):
    class_a = navigable_association.memberEnd[0].type

    attr_def = list(variables(class_a))

    assert class_a.name == "A"
    assert attr_def == ["b: relation_many[B]"]


def test_coder_write_class_with_1_n_association(navigable_association):
    UML.recipes.set_multiplicity_upper_value(navigable_association.memberEnd[1], 1)
    class_a = navigable_association.memberEnd[0].type
    assert (
        UML.recipes.get_multiplicity_upper_value_as_string(
            navigable_association.memberEnd[1]
        )
        == "1"
    )

    attr_def = list(variables(class_a))

    assert class_a.name == "A"
    assert attr_def == ["b: relation_one[B]"]


def test_in_profile():
    class_ = UML.Class()
    profile = UML.Profile()
    profile.ownedType = class_

    assert is_in_profile(class_)


def test_not_in_profile():
    class_ = UML.Class()
    package = UML.Package()
    package.ownedType = class_

    assert not is_in_profile(class_)


def test_in_toplevel_package():
    class_ = UML.Class()
    package = UML.Package()
    nested = UML.Package()
    package.name = "Foo"
    package.nestedPackage = nested
    nested.ownedType = class_

    assert is_in_toplevel_package(class_, "Foo")
    assert not is_in_toplevel_package(class_, "Bar")


def by_name(name):
    return lambda e: isinstance(e, UML.Class) and e.name == name


def test_bases(uml_metamodel: ElementFactory):
    package = next(uml_metamodel.select(by_name("Package")))

    names = [s.name for s in bases(package)]

    assert "Namespace" in names
    assert "PackageableElement" in names


def test_extension_bases(element_factory: ElementFactory):
    metaclass = element_factory.create(UML.Class)
    stereotype = element_factory.create(UML.Stereotype)
    UML.recipes.create_extension(metaclass, stereotype)

    supers = list(bases(stereotype))

    assert supers == [metaclass]


def test_simple_attribute(uml_metamodel: ElementFactory):
    package = next(uml_metamodel.select(by_name("Package")))
    value_spec = next(uml_metamodel.select(by_name("ValueSpecification")))
    literal_spec = next(uml_metamodel.select(by_name("LiteralSpecification")))

    assert not is_simple_type(package)
    assert not is_simple_type(value_spec)
    assert not is_simple_type(literal_spec)


def test_order_classes(uml_metamodel):
    classes = list(order_classes(uml_metamodel.select(UML.Class)))

    assert classes[0].name == "Base"
    assert classes[1].name == "Element"
    assert classes[2].name == "NamedElement"


def test_coder_write_association(navigable_association: UML.Association):
    a = list(associations(navigable_association.memberEnd[0].type))

    assert a == ['A.b = association("b", B, opposite="a")']


def test_coder_write_association_lower_value(navigable_association: UML.Association):
    end = navigable_association.memberEnd[1]
    UML.recipes.set_multiplicity_lower_value(end, 1)

    a = list(associations(navigable_association.memberEnd[0].type))

    assert a == ['A.b = association("b", B, lower=1, opposite="a")']


def test_coder_write_association_upper_value(navigable_association: UML.Association):
    end = navigable_association.memberEnd[1]
    UML.recipes.set_multiplicity_upper_value(end, 1)

    a = list(associations(navigable_association.memberEnd[0].type))

    assert a == ['A.b = association("b", B, upper=1, opposite="a")']


def test_coder_write_association_composite(navigable_association: UML.Association):
    end = navigable_association.memberEnd[1]
    end.aggregation = "composite"

    a = list(associations(navigable_association.memberEnd[0].type))

    assert a == ['A.b = association("b", B, composite=True, opposite="a")']


def test_coder_write_association_not_navigable(navigable_association: UML.Association):
    UML.recipes.set_navigability(
        navigable_association, navigable_association.memberEnd[1], None
    )

    a = list(associations(navigable_association.memberEnd[0].type))

    assert not a


def test_coder_write_association_opposite_not_navigable(
    navigable_association: UML.Association,
):
    UML.recipes.set_navigability(
        navigable_association, navigable_association.memberEnd[0], None
    )

    a = list(associations(navigable_association.memberEnd[0].type))

    assert a == ['A.b = association("b", B)']


def test_attribute_from_super_model(
    uml_metamodel: ElementFactory, core_metamodel: ElementFactory
):
    package = UML.Package()
    package.name = "UML"
    class_ = UML.Class()
    class_.name = "Package"
    class_.owningPackage = package

    element_type, base = superset_attribute(
        class_,
        "member",
        {
            # Order matters! Base model first.
            "Core": (CoreModelingLanguage(), core_metamodel),
            "UML": (UMLModelingLanguage(), uml_metamodel),
        },
    )

    assert element_type is UML.Package
    assert base.owner.name == "Namespace"


def test_replace_simple_attribute(uml_metamodel: ElementFactory):
    klass = next(
        uml_metamodel.select(lambda e: isinstance(e, UML.Class) and e.name == "Class")
    )
    a = next(it for it in klass.ownedAttribute if it.name == "isActive")

    assert a.name == "isActive"
    assert a.typeValue == "bool"
    assert not a.type



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.classes.dependencypropertypages import (
    DependencyItemPropertyPage,
    DependencyPropertyPage,
)


def test_dependency_property_page(diagram, element_factory, event_manager):
    subject = element_factory.create(UML.Dependency)
    source = element_factory.create(UML.Class)
    source.name = "A"
    target = element_factory.create(UML.Class)
    target.name = "B"
    subject.client = source
    subject.supplier = target

    property_page = DependencyPropertyPage(subject, event_manager)

    widget = property_page.construct()
    head_title = find(widget, "head")
    tail_title = find(widget, "tail")
    head_text = head_title.get_text()
    tail_text = tail_title.get_text()

    assert head_text == "A"
    assert tail_text == "B"


def test_dependency_item_property_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.classes.DependencyItem, subject=element_factory.create(UML.Dependency)
    )
    property_page = DependencyItemPropertyPage(item, event_manager)

    widget = property_page.construct()
    dependency_combo = find(widget, "dependency-dropdown")
    dependency_combo.set_selected(2)

    assert item.dependency_type is UML.Realization


def test_dependency_item_property_page_without_subject(diagram, event_manager):
    item = diagram.create(UML.classes.DependencyItem)
    property_page = DependencyItemPropertyPage(item, event_manager)

    widget = property_page.construct()

    assert widget



from collections.abc import Iterable

from gaphor import UML
from gaphor.core.changeset.compare import compare
from gaphor.core.modeling import (
    ElementFactory,
    PendingChange,
)
from gaphor.diagram.tests.fixtures import connect
from gaphor.SysML import sysml
from gaphor.SysML.diagramitems import BlockItem, PropertyItem, ProxyPortItem
from gaphor.SysML.propertypages import create_item_flow
from gaphor.ui.modelmerge.organize import Node, organize_changes
from gaphor.UML.actions.actionstoolbox import partition_config
from gaphor.UML.diagramitems import (
    ActivityItem,
    AssociationItem,
    ClassItem,
    ConnectorItem,
    ExtensionItem,
    PartitionItem,
    StateItem,
    TransitionItem,
)
from gaphor.UML.recipes import apply_stereotype
from gaphor.UML.umllex import parse


def all_change_ids(nodes: Iterable[Node]):
    for n in nodes:
        yield from (e.id for e in n.elements)
        if n.children:
            yield from all_change_ids(n.children)


def test_class_with_members(element_factory, modeling_language, create):
    class_item = create(ClassItem, UML.Class)
    class_item.subject.name = "MyClass"
    class_item.subject.ownedAttribute = element_factory.create(UML.Property)
    parse(class_item.subject.ownedAttribute[0], "+foo: int")
    class_item.subject.ownedOperation = element_factory.create(UML.Operation)
    parse(class_item.subject.ownedOperation[0], "+ to_string(arg: int): str")

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 1
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_association(element_factory, modeling_language, create):
    head_class_item = create(ClassItem, UML.Class)
    tail_class_item = create(ClassItem, UML.Class)
    association = create(AssociationItem)
    connect(association, association.head, head_class_item)
    connect(association, association.tail, tail_class_item)

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 1
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_association_with_existing_classes(element_factory, modeling_language, create):
    head_class_item = create(ClassItem, UML.Class)
    tail_class_item = create(ClassItem, UML.Class)
    association = create(AssociationItem)
    connect(association, association.head, head_class_item)
    connect(association, association.tail, tail_class_item)

    current = ElementFactory()
    current.create_as(UML.Class, head_class_item.subject.id)
    current.create_as(UML.Class, tail_class_item.subject.id)
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 3
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_extension(element_factory, modeling_language, create):
    class_item = create(ClassItem, UML.Class)
    stereotype_item = create(ClassItem, UML.Stereotype)
    extension = create(ExtensionItem)
    connect(extension, extension.head, class_item)
    connect(extension, extension.tail, stereotype_item)

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 1
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_activity_with_parameter_node(element_factory, modeling_language, create):
    activity = create(ActivityItem, UML.Activity)
    node = element_factory.create(UML.ActivityParameterNode)
    node.parameter = element_factory.create(UML.Parameter)
    activity.subject.node = node

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 1
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_partition_with_swim_lanes(element_factory, modeling_language, create):
    partition = create(PartitionItem, UML.ActivityPartition)
    # NB. This adds an extra Activity namespace!
    partition_config(partition)

    partition.partition = element_factory.create(UML.ActivityPartition)
    partition.partition = element_factory.create(UML.ActivityPartition)
    partition.partition[0].activity = partition.subject.activity
    partition.partition[1].activity = partition.subject.activity

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    # Create diagram + extra Activity
    assert len(tree) == 2
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_connector_with_item_flow(element_factory, modeling_language, create):
    block_item = create(BlockItem, sysml.Block)
    property_item = create(PropertyItem, UML.Property)
    proxy_port_item = create(ProxyPortItem)
    connector_item = create(ConnectorItem)
    connect(proxy_port_item, proxy_port_item.handles()[0], block_item)
    connect(connector_item, connector_item.head, proxy_port_item)
    connect(connector_item, connector_item.tail, property_item)
    create_item_flow(connector_item.subject)

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 3
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_state_with_entry_do_exit_criteria(element_factory, modeling_language, create):
    state_item = create(StateItem, UML.State)
    state_item.subject.entry = element_factory.create(UML.Activity)
    state_item.subject.doActivity = element_factory.create(UML.Activity)

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert len(tree) == 1
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_transition_with_guard(element_factory, modeling_language, create):
    head_state_item = create(StateItem, UML.State)
    tail_state_item = create(StateItem, UML.State)
    transition_item = create(TransitionItem)
    connect(transition_item, transition_item.head, head_state_item)
    connect(transition_item, transition_item.tail, tail_state_item)

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    assert transition_item.subject.guard
    assert len(tree) == 1
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))


def test_applied_stereotype(element_factory, modeling_language, create):
    stereotype = element_factory.create(UML.Stereotype)
    class_item = create(ClassItem, UML.Class)
    apply_stereotype(class_item.subject, stereotype)

    current = ElementFactory()
    all(compare(current, current, element_factory))
    tree = list(organize_changes(current, modeling_language))

    # Create Stereotype + create diagram
    assert len(tree) == 2
    assert {e.id for e in current.select(PendingChange)} == set(all_change_ids(tree))



"""Every now and then I run into an issue where a derived union returns a
unioncache.

It looks like there is a unioncache packed in a unioncache.

What makes the issue even more weird is that the derived field is not even assigned on this type.
"""

import pytest

from gaphor import UML
from gaphor.application import Session
from gaphor.core import Transaction
from gaphor.core.modeling import Diagram
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML import diagramitems


@pytest.fixture
def session():
    session = Session()
    yield session
    session.shutdown()


@pytest.fixture
def event_manager(session):
    return session.get_service("event_manager")


@pytest.fixture
def element_factory(session):
    return session.get_service("element_factory")


@pytest.fixture
def diagram(element_factory, event_manager):
    with Transaction(event_manager):
        return element_factory.create(Diagram)


def test_unioncache_in_derived_union(diagram, event_manager, element_factory):
    with Transaction(event_manager):
        uc1 = diagram.create(
            diagramitems.UseCaseItem, subject=element_factory.create(UML.UseCase)
        )
        uc2 = diagram.create(
            diagramitems.UseCaseItem, subject=element_factory.create(UML.UseCase)
        )
        include = diagram.create(diagramitems.IncludeItem)

        connect(include, include.handles()[0], uc1)
        connect(include, include.handles()[1], uc2)

    assert uc1.subject in include.subject.target
    assert include.subject.ownedElement == []



import pytest

from gaphor import UML
from gaphor.core.format import format, parse
from gaphor.diagram.copypaste import copy_full, paste_full
from gaphor.diagram.tests.fixtures import (
    connect,
    copy_and_paste_link,
    copy_clear_and_paste_link,
)
from gaphor.UML.classes import (
    AssociationItem,
    ClassItem,
    DataTypeItem,
    EnumerationItem,
    InterfaceItem,
)


@pytest.mark.parametrize(
    "cls_type,item",
    [
        (UML.Class, ClassItem),
        (UML.DataType, DataTypeItem),
        (UML.Enumeration, EnumerationItem),
    ],
)
def test_class_with_attributes(diagram, element_factory, cls_type, item):
    cls = element_factory.create(cls_type)
    attr = element_factory.create(UML.Property)
    parse(attr, "- attr: str")
    cls.ownedAttribute = attr

    cls_item = diagram.create(item, subject=cls)

    new_items = copy_clear_and_paste_link({cls_item}, diagram, element_factory)
    new_cls_item = new_items.pop()

    assert isinstance(new_cls_item, item)
    assert format(new_cls_item.subject.ownedAttribute[0]) == "- attr: str"


@pytest.mark.parametrize(
    "cls_type,item",
    [
        (UML.Class, ClassItem),
        (UML.DataType, DataTypeItem),
        (UML.Enumeration, EnumerationItem),
    ],
)
def test_class_with_operation(diagram, element_factory, cls_type, item):
    cls = element_factory.create(cls_type)
    oper = element_factory.create(UML.Operation)
    parse(oper, "- oper(inout param: str): str")
    cls.ownedOperation = oper

    cls_item = diagram.create(item, subject=cls)

    new_items = copy_clear_and_paste_link({cls_item}, diagram, element_factory)
    new_cls_item = new_items.pop()

    assert isinstance(new_cls_item, item)
    assert (
        format(new_cls_item.subject.ownedOperation[0])
        == "- oper(inout param: str): str"
    )


def test_enumeration_with_literal(diagram, element_factory):
    enum = element_factory.create(UML.Enumeration)
    literal = element_factory.create(UML.EnumerationLiteral)
    parse(literal, "apple")
    enum.ownedLiteral = literal

    enum_item = diagram.create(EnumerationItem, subject=enum)

    new_items = copy_clear_and_paste_link({enum_item}, diagram, element_factory)
    new_enum_item = new_items.pop()

    assert isinstance(new_enum_item, EnumerationItem)
    assert format(new_enum_item.subject.ownedLiteral[0]) == "apple"


def test_interface_with_attributes_and_operation(diagram, element_factory):
    iface = element_factory.create(UML.Interface)

    attr = element_factory.create(UML.Property)
    parse(attr, "- attr: str")
    iface.ownedAttribute = attr

    oper = element_factory.create(UML.Operation)
    parse(oper, "- oper(inout param: str): str")
    iface.ownedOperation = oper

    iface_item = diagram.create(InterfaceItem, subject=iface)

    new_items = copy_clear_and_paste_link({iface_item}, diagram, element_factory)
    new_iface_item = new_items.pop()

    assert isinstance(new_iface_item, InterfaceItem)
    assert format(new_iface_item.subject.ownedAttribute[0]) == "- attr: str"
    assert (
        format(new_iface_item.subject.ownedOperation[0])
        == "- oper(inout param: str): str"
    )


def two_classes_and_an_association(diagram, element_factory):
    gen_cls = element_factory.create(UML.Class)
    spc_cls = element_factory.create(UML.Class)
    gen_cls_item = diagram.create(ClassItem, subject=gen_cls)
    spc_cls_item = diagram.create(ClassItem, subject=spc_cls)
    assoc_item = diagram.create(AssociationItem)
    gen_cls.name = "Gen"
    spc_cls.name = "Spc"

    connect(assoc_item, assoc_item.handles()[0], spc_cls_item)
    connect(assoc_item, assoc_item.handles()[1], gen_cls_item)
    UML.recipes.set_navigability(
        assoc_item.subject, assoc_item.subject.memberEnd[0], True
    )

    assert (
        assoc_item.subject.memberEnd[0]
        in assoc_item.subject.memberEnd[1].type.ownedAttribute
    )
    # assert gen_cls_item.ownedAttribute

    return gen_cls_item, spc_cls_item, assoc_item


def test_copy_paste_items_with_connections(diagram, element_factory):
    gen_cls_item, spc_cls_item, assoc_item = two_classes_and_an_association(
        diagram, element_factory
    )

    assert assoc_item.head_subject
    assert assoc_item.tail_subject

    assoc = assoc_item.subject

    copy_and_paste_link(
        {gen_cls_item, assoc_item, spc_cls_item}, diagram, element_factory
    )

    new_assoc_item = assoc.presentation[1]

    assert assoc.presentation[0] is assoc_item

    assert new_assoc_item.head_subject
    assert new_assoc_item.tail_subject


def test_copy_remove_paste_items_with_connections(diagram, element_factory):
    gen_cls_item, spc_cls_item, assoc_item = two_classes_and_an_association(
        diagram, element_factory
    )

    new_items = copy_clear_and_paste_link(
        {gen_cls_item, assoc_item, spc_cls_item}, diagram, element_factory
    )
    new_cls1 = next(
        element_factory.select(
            lambda e: isinstance(e, UML.NamedElement) and e.name == "Gen"
        )
    )
    new_cls2 = next(
        element_factory.select(
            lambda e: isinstance(e, UML.NamedElement) and e.name == "Spc"
        )
    )
    new_assoc = next(element_factory.select(UML.Association))

    assert new_assoc.memberEnd[0].type is new_cls2
    assert new_assoc.memberEnd[1].type is new_cls1
    assert new_assoc.memberEnd[0] in new_assoc.memberEnd[1].type.ownedAttribute
    assert new_cls1.presentation[0] in new_items
    assert new_cls2.presentation[0] in new_items
    assert new_assoc.presentation[0] in new_items
    assert new_assoc.presentation[0].head_subject
    assert new_assoc.presentation[0].tail_subject


@pytest.mark.parametrize("uml_type", [UML.Property, UML.Parameter])
def test_copy_multiplicity(uml_type, diagram, element_factory):
    prop = element_factory.create(uml_type)
    prop.lowerValue = element_factory.create(UML.LiteralInteger)
    prop.lowerValue.value = 1
    prop.upperValue = element_factory.create(UML.LiteralInteger)
    prop.upperValue.value = 2

    buffer = copy_full({prop})
    prop.unlink()
    paste_full(buffer, diagram)

    new_prop = next(element_factory.select(uml_type))

    assert new_prop.lowerValue
    assert new_prop.lowerValue.value == 1
    assert new_prop.upperValue
    assert new_prop.upperValue.value == 2


def test_copy_property_default_value(diagram, element_factory):
    prop = element_factory.create(UML.Property)
    prop.defaultValue = element_factory.create(UML.LiteralString)
    prop.defaultValue.value = "foo"
    prop.lowerValue = element_factory.create(UML.LiteralInteger)
    prop.lowerValue.value = 1
    prop.upperValue = element_factory.create(UML.LiteralInteger)
    prop.upperValue.value = 2

    buffer = copy_full({prop})
    prop.unlink()
    paste_full(buffer, diagram)

    new_prop = next(element_factory.select(UML.Property))

    assert new_prop.defaultValue
    assert new_prop.defaultValue.value == "foo"



import pytest
import pytest_asyncio

from gaphor.application import Session
from gaphor.core import Transaction
from gaphor.core.modeling import Diagram
from gaphor.diagram.diagramtoolbox import new_item_factory
from gaphor.diagram.tools.placement import (
    PlacementState,
    on_drag_begin,
    on_drag_end,
    placement_tool,
)
from gaphor.ui.diagrampage import DiagramPage
from gaphor.UML.general import CommentLineItem


@pytest.fixture
def session():
    session = Session()
    yield session
    session.shutdown()


@pytest.fixture
def event_manager(session):
    return session.get_service("event_manager")


@pytest.fixture
def element_factory(session):
    return session.get_service("element_factory")


@pytest.fixture
def modeling_language(session):
    return session.get_service("modeling_language")


@pytest.fixture
def diagram(element_factory, event_manager):
    with Transaction(event_manager):
        return element_factory.create(Diagram)


@pytest_asyncio.fixture
async def view(diagram, event_manager, element_factory, modeling_language):
    page = DiagramPage(diagram, event_manager, element_factory, modeling_language)
    page.construct()
    return page.view


def click(tool, event_manager):
    factory = new_item_factory(CommentLineItem)
    state = PlacementState(factory, event_manager, handle_index=-1)
    on_drag_begin(tool, 0, 0, state)
    on_drag_end(tool, 0, 0, state)


def test_placement(view, event_manager):
    factory = new_item_factory(CommentLineItem)
    tool = placement_tool(factory, event_manager, handle_index=-1)
    view.add_controller(tool)

    with Transaction(event_manager):
        click(tool, event_manager)
        click(tool, event_manager)



import sys

from gaphor.services.moduleloader import ModuleLoader


def test_module_loader():
    ModuleLoader()

    assert "gaphor.diagram.general.uicomponents" in sys.modules
    assert "gaphor.UML.uicomponents" in sys.modules
    assert "gaphor.SysML.propertypages" in sys.modules
    assert "gaphor.C4Model.propertypages" in sys.modules



from pathlib import Path

import pytest
from gaphas.solver import VERY_STRONG, Solver, Variable

from gaphor import UML
from gaphor.UML.general.image import AspectRatioConstraint, ImageItem


def test_image_property_minimum_size(create):
    image = create(ImageItem, UML.Image)

    assert image.min_width == 10
    assert image.min_height == 10


def test_image_property_set_size(create):
    image = create(ImageItem, UML.Image)

    image.load_image_from_file(Path("data/logos/gaphor-24x24.png"))

    assert image.width == 24
    assert image.height == 24


@pytest.mark.parametrize(
    ["ratio", "set_index", "set_value", "expect_width", "expect_height"],
    [
        [1.0, 0, 50.0, 50.0, 50.0],
        [0.4, 0, 50.0, 50.0, 125.0],
        [1.5, 1, 50.0, 75.0, 50.0],
        [1.0, 4, 2.0, 200.0, 100.0],
        # This values look off, due to how the weakest variable is determined:
        [1.0, 2, 50.0, 100.0, 100.0],
        [0.4, 2, 50.0, 40.0, 100.0],
    ],
)
def test_aspect_ratio_change(ratio, set_index, set_value, expect_width, expect_height):
    x0, y0, x1, y1 = (Variable(value=v, strength=VERY_STRONG) for v in [0, 0, 100, 100])
    ratio = Variable(value=ratio, strength=VERY_STRONG)
    solver = Solver()
    vars = [x0, y0, x1, y1, ratio]
    constraint = AspectRatioConstraint(*vars)
    solver.add_constraint(constraint)

    vars[set_index].value = set_value
    solver.solve()

    assert x1 - x0 == pytest.approx(expect_width)
    assert y1 - y0 == pytest.approx(expect_height)



import pytest

from gaphor.diagram.export import (
    diagram_render_offsets,
    escape_filename,
    save_eps,
    save_pdf,
    save_png,
    save_svg,
)
from gaphor.diagram.general import Box


@pytest.fixture
def diagram_with_box(diagram):
    diagram.create(Box)
    return diagram


def test_export_to_svg(diagram_with_box, tmp_path):
    f = tmp_path / "test.svg"

    save_svg(f, diagram_with_box)
    content = f.read_text(encoding="utf-8")

    assert "<svg" in content


def test_export_to_png(diagram_with_box, tmp_path):
    f = tmp_path / "test.png"

    save_png(f, diagram_with_box)
    content = f.read_bytes()

    assert b"PNG" in content


def test_export_to_pdf(diagram_with_box, tmp_path):
    f = tmp_path / "test.pdf"

    save_pdf(f, diagram_with_box)
    content = f.read_bytes()

    assert b"%PDF" in content


def test_export_to_eps(diagram_with_box, tmp_path):
    f = tmp_path / "test.eps"

    save_eps(f, diagram_with_box)
    content = f.read_bytes()

    assert b"%!PS-Adobe-3.0 EPSF-3.0" in content


def test_diagram_render_offsets_respects_padding(diagram_with_box):
    tx8, ty8 = diagram_render_offsets(diagram_with_box, padding=8)
    tx20, ty20 = diagram_render_offsets(diagram_with_box, padding=20)
    assert (tx20 - tx8) == pytest.approx(12.0)
    assert (ty20 - ty8) == pytest.approx(12.0)


def test_escape_filename():
    assert escape_filename("foo bar") == "foo_bar"
    assert escape_filename(r"foo \ bar >") == "foo_bar_"
    assert escape_filename("çëÆØ") == "çëÆØ"
    assert escape_filename("こんにちは") == "こんにちは"  # should read: "hello"



import pytest

from gaphor import UML
from gaphor.diagram.tests.fixtures import connect
from gaphor.plugins.autolayout.pydot import AutoLayout, parse_edge_pos, strip_quotes
from gaphor.UML.diagramitems import (
    ActionItem,
    AssociationItem,
    ClassItem,
    ForkNodeItem,
    GeneralizationItem,
    InputPinItem,
    ObjectFlowItem,
    PackageItem,
)
from gaphor.UML.general import CommentItem, CommentLineItem


def test_layout_diagram(diagram, create):
    superclass = create(ClassItem, UML.Class)
    subclass = create(ClassItem, UML.Class)
    gen = create(GeneralizationItem, UML.Generalization)
    connect(gen, gen.tail, superclass)
    connect(gen, gen.head, subclass)

    auto_layout = AutoLayout()
    auto_layout.layout(diagram)

    assert gen.head.pos != (0, 0)
    assert gen.tail.pos != (0, 0)


def test_layout_with_association(diagram, create, event_manager):
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    a = create(AssociationItem)
    connect(a, a.head, c1)
    connect(a, a.tail, c2)

    auto_layout = AutoLayout(event_manager)
    auto_layout.layout(diagram)


def test_layout_with_comment(diagram, create, event_manager):
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    a = create(AssociationItem)
    connect(a, a.head, c1)
    connect(a, a.tail, c2)

    comment = create(CommentItem, UML.Comment)
    comment_line = create(CommentLineItem)
    connect(comment_line, comment_line.head, comment)
    connect(comment_line, comment_line.tail, a)

    auto_layout = AutoLayout(event_manager)
    auto_layout.layout(diagram)


def test_layout_with_nested(diagram, create, event_manager):
    p = create(PackageItem, UML.Package)
    c1 = create(ClassItem, UML.Class)
    p.children = c1
    c2 = create(ClassItem, UML.Class)
    a = create(AssociationItem)
    connect(a, a.head, c1)
    connect(a, a.tail, c2)

    auto_layout = AutoLayout(event_manager)
    auto_layout.layout(diagram)

    assert c1.matrix[4] < p.width
    assert c1.matrix[5] < p.height


def test_layout_with_attached_item(diagram, create, event_manager):
    action = create(ActionItem, UML.Action)
    pin = create(InputPinItem, UML.InputPin)
    connect(pin, pin.handles()[0], action)

    action2 = create(ActionItem, UML.Action)
    object_flow = create(ObjectFlowItem, UML.ObjectFlow)
    connect(object_flow, object_flow.head, pin)
    connect(object_flow, object_flow.tail, action2)

    auto_layout = AutoLayout(event_manager)
    auto_layout.layout(diagram)

    assert pin.parent is action


def test_layout_fork_node_item(diagram, create, event_manager):
    create(ForkNodeItem, UML.ForkNode)

    auto_layout = AutoLayout(event_manager)
    auto_layout.layout(diagram)


def test_parse_pos():
    points = parse_edge_pos('"1.0,2.0 3.0,4 5.0,6.0 7,8.0"', 10, True)

    assert points == [(7.0, 2.0), (1.0, 8.0)]


def test_parse_pos_invalid_number_of_points():
    with pytest.raises(IndexError):
        parse_edge_pos('"1.0,2.0 3.0,4 5.0,6.0"', 10, True)


def test_strip_line_endings():
    assert strip_quotes("\\\n807.5") == "807.5"
    assert strip_quotes("\\\r\n807.5") == "807.5"
    assert strip_quotes('\\\r\n"807.5"') == "807.5"



from gaphor import UML
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.classes.association import AssociationItem
from gaphor.UML.classes.datatype import DataTypeItem
from gaphor.UML.classes.klass import ClassItem
from gaphor.UML.recipes import set_navigability


def test_connect_value_type_to_block(diagram, element_factory):
    b = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    v = diagram.create(DataTypeItem, subject=element_factory.create(UML.DataType))
    a = diagram.create(AssociationItem)

    connect(a, a.head, b)
    connect(a, a.tail, v)
    set_navigability(a.subject, a.head_end.subject, True)
    set_navigability(a.subject, a.tail_end.subject, True)
    a.tail_end.subject.aggregation = "composite"

    assert a.head_end.subject.type is b.subject
    assert a.tail_end.subject.type is v.subject
    assert a.tail_end.subject in b.subject.ownedAttribute



from gaphor import UML
from gaphor.UML.iconname import get_name_for_pseudostate


def test_pseudostate_icon():
    element = UML.Pseudostate()
    element.kind = "initial"

    assert get_name_for_pseudostate(element) == "gaphor-initial-pseudostate-symbolic"



from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.tests.fixtures import allow, connect, get_connected
from gaphor.UML.classes.generalization import GeneralizationItem
from gaphor.UML.classes.klass import ClassItem


def test_generalization_multiple_connection(create, element_factory):
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    gen1 = create(GeneralizationItem)

    connect(gen1, gen1.tail, c1)
    connect(gen1, gen1.head, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    gen2 = diagram2.create(GeneralizationItem)

    connect(gen2, gen2.tail, c3)
    connect(gen2, gen2.head, c4)

    assert gen1.subject is gen2.subject


def test_generalization_glue(create):
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    glued = allow(gen, gen.tail, c1)
    assert glued

    connect(gen, gen.tail, c1)
    assert get_connected(gen, gen.tail) is c1
    assert gen.subject is None

    glued = allow(gen, gen.head, c2)
    assert glued


def test_generalization_connect(create):
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.head, c1)
    connect(gen, gen.tail, c2)

    assert get_connected(gen, gen.head) is c1
    assert get_connected(gen, gen.tail) is c2
    assert gen.subject
    assert gen.subject.specific is c1.subject
    assert gen.subject.general is c2.subject


def test_generalization_connect_in_new_diagram(create, element_factory):
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.head, c1)
    connect(gen, gen.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    gen2 = diagram2.create(GeneralizationItem)

    connect(gen2, gen2.head, c3)
    connect(gen2, gen2.tail, c4)
    assert gen.subject is gen2.subject


def test_generalization_connect_in_new_diagram_in_opposite_direction(
    create, element_factory
):
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.tail, c1)
    connect(gen, gen.head, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    gen2 = diagram2.create(GeneralizationItem)

    connect(gen2, gen2.head, c3)
    connect(gen2, gen2.tail, c4)
    assert gen.subject is not gen2.subject
    assert gen.subject.specific is c2.subject
    assert gen.subject.general is c1.subject
    assert gen2.subject.specific is c3.subject
    assert gen2.subject.general is c4.subject


def test_generalization_reconnect_in_same_diagram(create):
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    c3 = create(ClassItem, UML.Class)
    gen = create(GeneralizationItem)

    # connect: c1 -> c2
    connect(gen, gen.head, c1)
    connect(gen, gen.tail, c2)

    s = gen.subject

    # reconnect: c2 -> c3
    connect(gen, gen.tail, c3)

    assert s is not gen.subject
    assert c1.subject is gen.subject.specific
    assert c3.subject is gen.subject.general
    assert c2.subject is not gen.subject.general


def test_generalization_reconnect_in_new_diagram(create, element_factory):
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.head, c1)
    connect(gen, gen.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    gen2 = diagram2.create(GeneralizationItem)

    connect(gen2, gen2.head, c3)
    connect(gen2, gen2.tail, c4)
    assert gen.subject is gen2.subject

    c5 = diagram2.create(ClassItem, subject=element_factory.create(UML.Class))
    connect(gen2, gen2.head, c5)

    assert gen.subject is not gen2.subject
    assert gen.subject.specific is c1.subject
    assert gen.subject.general is c2.subject
    assert gen2.subject.specific is c5.subject
    assert gen2.subject.general is c4.subject


def test_genralization_reconnect_twice_in_new_diagram(create, element_factory):
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.head, c1)
    connect(gen, gen.tail, c2)

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    gen2 = diagram2.create(GeneralizationItem)

    connect(gen2, gen2.head, c3)
    connect(gen2, gen2.tail, c4)
    assert gen.subject is gen2.subject

    c5 = diagram2.create(ClassItem, subject=c2.subject)
    connect(gen2, gen2.head, c5)
    connect(gen2, gen2.head, c3)

    assert gen.subject is gen2.subject
    assert gen.subject.specific is c3.subject
    assert gen.subject.general is c4.subject



import sys
import textwrap

import pytest
from dulwich.repo import Repo

from gaphor import UML
from gaphor.storage.tests.fixtures import create_merge_conflict
from gaphor.ui.filemanager import FileManager


@pytest.fixture
def file_manager(event_manager, element_factory, modeling_language):
    main_window = None
    file_manager = FileManager(
        event_manager, element_factory, modeling_language, main_window
    )
    yield file_manager
    file_manager.shutdown()


@pytest.mark.asyncio
async def test_save(element_factory, file_manager: FileManager, tmp_path):
    element_factory.create(UML.Class)
    out_file = tmp_path / "out.gaphor"

    await file_manager.save(filename=out_file)

    assert out_file.exists()


@pytest.mark.asyncio
async def test_model_is_saved_with_utf8_encoding(
    element_factory, file_manager: FileManager, tmp_path
):
    class_ = element_factory.create(UML.Class)
    class_.name = "üëïèàòù"
    package = element_factory.create(UML.Package)
    package.name = "안녕하세요 세계"

    model_file = tmp_path / "model.gaphor"
    await file_manager.save(model_file)

    with open(model_file, encoding="utf-8") as f:
        f.read()  # raises exception if characters can't be decoded


@pytest.mark.asyncio
async def test_model_is_loaded_with_utf8_encoding(
    element_factory, file_manager: FileManager, tmp_path
):
    class_name = "üëïèàòù"
    package_name = "안녕하세요 세계"

    class_ = element_factory.create(UML.Class)
    class_.name = class_name
    package = element_factory.create(UML.Package)
    package.name = package_name

    model_file = tmp_path / "model.gaphor"
    await file_manager.save(model_file)

    element_factory.flush()

    await file_manager.load(model_file)
    new_class = next(element_factory.select(UML.Class))
    new_package = next(element_factory.select(UML.Package))

    assert new_class.name == class_name
    assert new_package.name == package_name


@pytest.mark.skipif(
    sys.platform != "win32", reason="Standard encoding on Windows is not UTF-8"
)
@pytest.mark.asyncio
async def test_old_model_is_loaded_without_utf8_encoding(
    file_manager: FileManager, test_models
):
    model_file = test_models / "wrong-encoding.gaphor"
    await file_manager.load(model_file)


@pytest.mark.asyncio
@pytest.mark.parametrize("resolution", ["current", "incoming"])
@pytest.mark.filterwarnings("ignore:use .* Repo._get_user_identity:DeprecationWarning")
async def test_load_model_with_merge_conflict(
    file_manager: FileManager, element_factory, merge_conflict, monkeypatch, resolution
):
    replace_merge_conflict_dialog(monkeypatch, resolution)

    await file_manager.resolve_merge_conflict(merge_conflict)

    assert element_factory.size() > 0


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:use .* Repo._get_user_identity:DeprecationWarning")
async def test_load_model_merge_conflict_and_manual_resolution(
    file_manager: FileManager, element_factory, merge_conflict, monkeypatch
):
    replace_merge_conflict_dialog(monkeypatch, "manual")

    await file_manager.resolve_merge_conflict(merge_conflict)

    from gaphor.core.modeling import PendingChange

    assert element_factory.lselect(PendingChange)


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:use .* Repo._get_user_identity:DeprecationWarning")
async def test_load_model_with_merge_conflict_and_unknown_resolution(
    file_manager: FileManager, merge_conflict, monkeypatch
):
    replace_merge_conflict_dialog(monkeypatch, "nonsense")

    with pytest.raises(ValueError):
        await file_manager.resolve_merge_conflict(merge_conflict)


def replace_merge_conflict_dialog(monkeypatch, resolution):
    async def mock_merge_conflict_dialog(_window):
        return resolution

    monkeypatch.setattr(
        "gaphor.ui.filemanager.resolve_merge_conflict_dialog",
        mock_merge_conflict_dialog,
    )


@pytest.fixture
def merge_conflict(tmp_path):
    initial_model = textwrap.dedent(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.12.1">
        <StyleSheet id="58d6989a-66f8-11ec-b4c8-0456e5e540ed"/>
        <Package id="58d6c2e8-66f8-11ec-b4c8-0456e5e540ed">
        <name>
        <val>current</val>
        </name>
        </Package>
        </gaphor>
        """
    )

    current_model = textwrap.dedent(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.12.1">
        <StyleSheet id="58d6989a-66f8-11ec-b4c8-0456e5e540ed"/>
        <Package id="58d6c2e8-66f8-11ec-b4c8-0456e5e540ed">
        <name>
        <val>current</val>
        </name>
        <ownedDiagram>
        <reflist>
        <ref refid="58d6c536-66f8-11ec-b4c8-0456e5e540ed"/>
        </reflist>
        </ownedDiagram>
        </Package>
        <Diagram id="58d6c536-66f8-11ec-b4c8-0456e5e540ed">
        <element>
        <ref refid="58d6c2e8-66f8-11ec-b4c8-0456e5e540ed"/>
        </element>
        <name>
        <val>diagram</val>
        </name>
        </Diagram>
        </gaphor>"""
    )

    incoming_model = textwrap.dedent(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.12.1">
        <StyleSheet id="58d6989a-66f8-11ec-b4c8-0456e5e540ed"/>
        <Package id="58d6c2e8-66f8-11ec-b4c8-0456e5e540ed">
        <name>
        <val>incoming</val>
        </name>
        </Package>
        </gaphor>"""
    )

    model = tmp_path / "model.gaphor"
    repo = Repo.init(tmp_path)

    create_merge_conflict(repo, model, initial_model, current_model, incoming_model)

    return model



import pytest

from gaphor.services.modelinglanguage import ModelingLanguageService


@pytest.fixture
def modeling_language(event_manager):
    return ModelingLanguageService(event_manager=event_manager, properties={})


def test_loading_of_modeling_languages(modeling_language):
    ids = [id for id, name in modeling_language.modeling_languages]
    assert "UML" in ids


def test_lookup_uml_element(modeling_language):
    assert modeling_language.lookup_element("Class")


def test_lookup_sysml_element(modeling_language):
    assert modeling_language.lookup_element("Block")


def test_lookup_c4model_element(modeling_language):
    assert modeling_language.lookup_element("Database")



# ruff: noqa: T201
"""Babel gettext helpers for Gaphor."""

import json
import subprocess
import sys
import urllib.request
from pathlib import Path

po_path: Path = Path(__file__).resolve().parent.parent / "po"
RELEASE_MATURE_LOCALES_FILE = po_path / "release_mature_locales.json"

WEBLATE_TRANSLATIONS_URL = (
    "https://hosted.weblate.org/api/components/gaphor/gaphor/translations/"
    "?page_size=100"
)


def run_babel(command: str, input: Path, output_file: Path, locale: str):
    subprocess.run(
        [
            "pybabel",
            command,
            f"--input={input}",
            f"--output-file={output_file}",
            f"--locale={locale}",
            "--domain=gaphor",
        ],
        check=True,
    )


def update_po_files():
    pot_path = po_path / "gaphor.pot"
    for path in (path for path in po_path.iterdir() if path.suffix == ".po"):
        run_babel("update", pot_path, path, path.stem)


def compile_mo_file(path: Path):
    mo_path = (
        po_path.parent / "gaphor" / "locale" / path.stem / "LC_MESSAGES" / "gaphor.mo"
    )
    mo_path.parent.mkdir(parents=True, exist_ok=True)
    run_babel("compile", path, mo_path, path.stem)


def compile_mo_all():
    for path in (path for path in po_path.iterdir() if path.suffix == ".po"):
        compile_mo_file(path)


def _fetch_mature_locales(first_url: str, min_percent: float) -> set[str]:
    """Return language codes at or above *min_percent* from the Weblate API.

    Exits the process with status 1 on network or parse errors.
    """
    codes: set[str] = set()
    url: str | None = first_url
    try:
        while url:
            req = urllib.request.Request(url, headers={"Accept": "application/json"})
            with urllib.request.urlopen(req, timeout=30) as resp:
                payload = json.load(resp)
            for row in payload.get("results", []):
                if row.get("is_source"):
                    continue
                code: str = row.get("language_code") or ""
                if code and float(row.get("translated_percent") or 0.0) >= min_percent:
                    codes.add(code)
            url = payload.get("next") or None
    except (OSError, json.JSONDecodeError) as e:
        print(f"Error: Weblate API request failed: {e}", file=sys.stderr)
        raise SystemExit(1) from None
    if not codes:
        print("Error: Weblate returned no mature locales.", file=sys.stderr)
        raise SystemExit(1)
    return codes


def update_release_mature_locales_file() -> None:
    """Fetch Weblate stats and write ``po/release_mature_locales.json``.

    Called by the monthly ``release-mature-locales-updater`` workflow
    (``poetry run poe release-mature-locales-update``).

    To change the threshold, edit ``min_translated_percent`` in
    ``po/release_mature_locales.json`` and commit the change.
    """
    data = json.loads(RELEASE_MATURE_LOCALES_FILE.read_text(encoding="utf-8"))
    min_pct = float(data.get("min_translated_percent", 72.0))

    from_api = _fetch_mature_locales(WEBLATE_TRANSLATIONS_URL, min_pct)
    po_stems = {p.stem for p in po_path.iterdir() if p.suffix == ".po"}
    chosen = sorted(from_api & po_stems)
    if not chosen:
        print(
            "Error: No intersection between Weblate mature locales and po/*.po.",
            file=sys.stderr,
        )
        raise SystemExit(1)

    new_text = (
        json.dumps({"min_translated_percent": min_pct, "locales": chosen}, indent=2)
        + "\n"
    )
    if RELEASE_MATURE_LOCALES_FILE.read_text(encoding="utf-8") == new_text:
        print("No change to po/release_mature_locales.json")
        return
    RELEASE_MATURE_LOCALES_FILE.write_text(new_text, encoding="utf-8")
    print(f"Wrote {RELEASE_MATURE_LOCALES_FILE} with {len(chosen)} locale(s).")


def release_language_codes() -> frozenset[str]:
    """Locales to compile for official release binaries.

    Read from ``po/release_mature_locales.json``, which is updated by the
    monthly ``release-mature-locales-updater`` workflow via
    ``poetry run poe release-mature-locales-update``.
    """
    data = json.loads(RELEASE_MATURE_LOCALES_FILE.read_text(encoding="utf-8"))
    po_stems = {p.stem for p in po_path.iterdir() if p.suffix == ".po"}
    return frozenset(data["locales"]) & po_stems


def compile_mo_release():
    mature_translations = release_language_codes()
    for path in po_path.iterdir():
        if path.suffix == ".po" and path.stem in mature_translations:
            compile_mo_file(path)


if __name__ == "__main__":
    update_po_files()
    compile_mo_all()



import pytest

from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.SysML.allocations.relationships import AllocateItem
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.interfaceblock import InterfaceBlockItem
from gaphor.SysML.sysml import Block, InterfaceBlock
from gaphor.UML.actions.activity import ActivityItem
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.states.statemachine import StateMachineItem
from gaphor.UML.uml import Activity, Interaction, StateMachine

# Some selection of items
items = [
    (Block, BlockItem),
    (InterfaceBlock, InterfaceBlockItem),
    (Activity, ActivityItem),
    (Interaction, InteractionItem),
    (StateMachine, StateMachineItem),
]


@pytest.mark.parametrize("source_types", items)
@pytest.mark.parametrize("target_types", items)
def test_allocate_relationship(create, diagram, source_types, target_types):
    source = create(source_types[1], source_types[0])
    target = create(target_types[1], target_types[0])

    allocate = diagram.create(AllocateItem)

    assert allow(allocate, allocate.handles()[0], source)
    assert allow(allocate, allocate.handles()[1], target)

    connect(allocate, allocate.handles()[0], source)
    connect(allocate, allocate.handles()[1], target)

    assert allocate.subject
    assert allocate.subject.sourceContext is source.subject
    assert allocate.subject.targetContext is target.subject

    disconnect(allocate, allocate.handles()[0])

    assert not allocate.subject



import pytest

from gaphor.core.modeling.modelinglanguage import CoreModelingLanguage


def test_name():
    modeling_language = CoreModelingLanguage()

    assert modeling_language.name == ""


def test_lookup_element():
    modeling_language = CoreModelingLanguage()

    assert modeling_language.lookup_element("Diagram")
    assert modeling_language.lookup_element("Presentation")
    assert modeling_language.lookup_element("StyleSheet")
    assert not modeling_language.lookup_element("NonExistent")


def test_toolbox():
    modeling_language = CoreModelingLanguage()

    with pytest.raises(ValueError):
        modeling_language.toolbox_definition  # noqa: B018



import pytest

from gaphor.core.modeling import Diagram
from gaphor.services.modelinglanguage import ModelingLanguageService
from gaphor.ui.event import CurrentDiagramChanged
from gaphor.ui.toolbox import Toolbox


@pytest.fixture
def modeling_language(event_manager):
    return ModelingLanguageService(event_manager)


@pytest.fixture
def toolbox(event_manager, properties, modeling_language):
    toolbox = Toolbox(
        event_manager=event_manager,
        properties=properties,
        modeling_language=modeling_language,
    )
    toolbox.open()
    yield toolbox
    toolbox.close()


def test_toolbox_construction(toolbox):
    widget = toolbox.open()

    assert widget


def test_current_tool(toolbox):
    tool_name = toolbox.active_tool_name

    assert tool_name == "toolbox-pointer"


def test_expanded_sections_for_default_diagram(element_factory, toolbox, event_manager):
    diagram = element_factory.create(Diagram)
    event_manager.handle(CurrentDiagramChanged(diagram))

    assert toolbox.expanded_sections() == {
        0: True,
        1: True,
        2: True,
        3: True,
        4: True,
        5: True,
        6: True,
        7: True,
    }


def text_expanded_sections_for_diagram_type(element_factory, toolbox, event_manager):
    diagram = element_factory.create(Diagram)
    diagram.diagramType = "sd"
    event_manager.handle(CurrentDiagramChanged(diagram))

    assert toolbox.expanded_sections() == {
        0: True,
        1: False,
        2: False,
        3: False,
        4: True,
        5: False,
        6: False,
        7: False,
    }


def test_expanded_sections_for_diagram_type_with_custom_settings(
    element_factory, toolbox, event_manager, properties
):
    diagram = element_factory.create(Diagram)
    diagram.diagramType = "sd"
    properties["toolbox-UML-sd-expanded"] = {2: True, 3: True}

    event_manager.handle(CurrentDiagramChanged(diagram))

    assert toolbox.expanded_sections() == {
        0: True,
        1: False,
        2: True,
        3: True,
        4: True,
        5: False,
        6: False,
        7: False,
    }


def test_expanded_sections_switch_from_generic_to_sequence_diagram(
    element_factory, toolbox, event_manager, properties
):
    generic = element_factory.create(Diagram)
    sequence = element_factory.create(Diagram)
    sequence.diagramType = "sd"
    properties["toolbox-UML-sd-expanded"] = {2: True, 3: True}
    event_manager.handle(CurrentDiagramChanged(generic))

    assert toolbox.expanded_sections() == {
        0: True,
        1: True,
        2: True,
        3: True,
        4: True,
        5: True,
        6: True,
        7: True,
    }

    event_manager.handle(CurrentDiagramChanged(sequence))

    assert toolbox.expanded_sections() == {
        0: True,
        1: False,
        2: True,
        3: True,
        4: True,
        5: False,
        6: False,
        7: False,
    }


# test custom expand a column, switch from diagram
def test_expanded_sections_switch_from_generic_to_custom_diagram(
    element_factory, toolbox, event_manager
):
    generic = element_factory.create(Diagram)
    sequence = element_factory.create(Diagram)
    sequence.diagramType = "sd"
    event_manager.handle(CurrentDiagramChanged(generic))

    assert toolbox.expanded_sections() == {
        0: True,
        1: True,
        2: True,
        3: True,
        4: True,
        5: True,
        6: True,
        7: True,
    }

    event_manager.handle(CurrentDiagramChanged(sequence))

    assert toolbox.expanded_sections() == {
        0: True,
        1: False,
        2: False,
        3: False,
        4: True,
        5: False,
        6: False,
        7: False,
    }



from dataclasses import replace

import cairo
import pytest
from gaphas.geometry import Rectangle

from gaphor.core.styling import (
    CompiledStyleSheet,
    JustifyContent,
    Number,
    Style,
    merge_styles,
)
from gaphor.diagram.shapes import (
    Box,
    CssNode,
    DrawContext,
    IconBox,
    Orientation,
    Shape,
    Text,
    TextAlign,
    UpdateContext,
    VerticalAlign,
    traverse_css_nodes,
)

FALLBACK_STYLE: Style = {
    "color": (0, 0, 0, 1),
    "font-family": "sans",
    "font-size": 14,
}


class InlineStyle:
    """Add inline styles for testing."""

    def __init__(
        self,
        inline_style: Style,
        child: Shape,
    ):
        self.inline_style = inline_style
        self.child = child

    def size(
        self, context: UpdateContext, bounding_box: Rectangle | None = None
    ) -> tuple[Number, Number]:
        style = merge_styles(context.style, self.inline_style)
        new_context = replace(context, style=style)

        return self.child.size(new_context, bounding_box)

    def draw(self, context: DrawContext, bounding_box: Rectangle):
        style = merge_styles(context.style, self.inline_style)
        new_context = replace(context, style=style)

        self.child.draw(new_context, bounding_box)


@pytest.fixture
def fixed_text_size(monkeypatch):
    size = (60, 15)

    def text_size(*args):
        return size

    monkeypatch.setattr("gaphor.diagram.text.Layout.size", text_size)
    return size


@pytest.fixture
def update_context():
    return UpdateContext(style=FALLBACK_STYLE)


@pytest.fixture
def draw_context():
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
    cr = cairo.Context(surface)
    return DrawContext(
        cairo=cr,
        style=FALLBACK_STYLE,
        selected=False,
        focused=False,
        hovered=False,
        dropzone=False,
    )


def test_box_size(update_context):
    box = Box()

    assert box.size(context=update_context) == (0, 0)


def test_draw_empty_box(draw_context):
    box = Box(draw=None)

    box.draw(context=draw_context, bounding_box=Rectangle())


def test_draw_box_with_custom_draw_function(draw_context):
    called = False

    def draw(box, context, bounding_box):
        nonlocal called
        called = True

    box = Box(draw=draw)

    box.draw(context=draw_context, bounding_box=Rectangle())

    assert called


def test_draw_last_box_with_all_remaining_space(update_context, draw_context):
    bounding_boxes = []

    def draw(_box, _context, bounding_box):
        bounding_boxes.append(bounding_box)

    box = InlineStyle(
        {"vertical-align": VerticalAlign.TOP},
        Box(
            InlineStyle({"min-height": 80}, Box(draw=draw)),
            Box(draw=draw),
        ),
    )

    box.size(context=update_context, bounding_box=Rectangle(0, 0, 100, 120))
    box.draw(context=draw_context, bounding_box=Rectangle(0, 0, 100, 120))

    assert bounding_boxes[0].height == 80
    assert bounding_boxes[1].height == 40


def test_draw_box_with_horzontal_content(update_context, draw_context):
    bounding_boxes = []

    def draw(_box, _context, bounding_box):
        bounding_boxes.append(bounding_box)

    box = Box(
        InlineStyle({"min-width": 40}, Box(draw=draw)),
        InlineStyle({"min-width": 80}, Box(draw=draw)),
        orientation=Orientation.HORIZONTAL,
    )

    box.size(context=update_context, bounding_box=Rectangle(0, 0, 120, 100))
    box.draw(context=draw_context, bounding_box=Rectangle(0, 0, 120, 100))

    assert bounding_boxes[0].width == 40
    assert bounding_boxes[0].height == 100
    assert bounding_boxes[1].width == 80
    assert bounding_boxes[1].height == 100


def test_draw_box_with_stretched_content(update_context, draw_context):
    bounding_boxes = []

    def draw(_box, _context, bounding_box):
        bounding_boxes.append(bounding_box)

    box = InlineStyle(
        {"justify-content": JustifyContent.STRETCH},
        Box(
            Box(draw=draw),
            Box(draw=draw),
        ),
    )

    box.size(context=update_context, bounding_box=Rectangle(0, 0, 100, 120))
    box.draw(context=draw_context, bounding_box=Rectangle(0, 0, 100, 120))

    assert bounding_boxes[0].height == 60
    assert bounding_boxes[1].height == 60


def test_draw_box_with_stretched_oversized_content(draw_context, update_context):
    bounding_boxes = []

    def draw(_box, _context, bounding_box):
        bounding_boxes.append(bounding_box)

    box = InlineStyle(
        {"justify-content": JustifyContent.STRETCH},
        Box(
            Box(draw=draw),
            InlineStyle({"min-height": 80}, Box(draw=draw)),
        ),
    )

    box.size(context=update_context, bounding_box=Rectangle(0, 0, 100, 120))
    box.draw(context=draw_context, bounding_box=Rectangle(0, 0, 100, 120))

    assert bounding_boxes[0].y == 0
    assert bounding_boxes[0].height == 40
    assert bounding_boxes[1].y == 40
    assert bounding_boxes[1].height == 80


def test_draw_icon_box(draw_context):
    box_drawn = None

    def box_draw(box, context, bounding_box):
        nonlocal box_drawn
        box_drawn = bounding_box

    shape = IconBox(Box(draw=box_draw), Text(text="some text"))

    bounding_box = Rectangle(11, 12, 13, 14)
    shape.draw(draw_context, bounding_box)

    assert box_drawn == bounding_box


def test_icon_box_child_placement_center_bottom(update_context):
    style = {"text-align": TextAlign.CENTER, "vertical-align": VerticalAlign.BOTTOM}

    text = Text(text="some text")
    shape = IconBox(
        Box(),
        text,
    )
    shape.size(update_context)

    w, _h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == (bounding_box.width - w) / 2
    assert y == bounding_box.height


def test_icon_box_child_placement_right_middle(update_context):
    style = {"text-align": TextAlign.RIGHT, "vertical-align": VerticalAlign.MIDDLE}

    text = Text(text="some text")
    shape = IconBox(
        Box(),
        text,
    )
    shape.size(update_context)

    _w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == bounding_box.width
    assert y == (bounding_box.height - h) / 2


def test_icon_box_child_placement_left_middle(update_context):
    style = {"text-align": TextAlign.LEFT, "vertical-align": VerticalAlign.MIDDLE}

    text = Text(text="some text")
    shape = IconBox(
        Box(),
        text,
    )
    shape.size(update_context)

    w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == -w
    assert y == (bounding_box.height - h) / 2


def test_icon_box_child_placement_center_top(update_context):
    style = {"text-align": TextAlign.CENTER, "vertical-align": VerticalAlign.TOP}

    text = Text(text="some text")
    shape = IconBox(
        Box(),
        text,
    )
    shape.size(update_context)

    w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == (bounding_box.width - w) / 2
    assert y == -h


def test_text_has_width(update_context, fixed_text_size):
    text = Text(lambda: "some text")

    w, _ = text.size(update_context)
    assert w == fixed_text_size[0]


def test_text_has_height(update_context, fixed_text_size):
    text = Text("some text")

    _, h = text.size(update_context)
    assert h == fixed_text_size[1]


def test_text_with_min_width(update_context):
    style = {"min-width": 100, "min-height": 0}
    text = InlineStyle(style, Text("some text"))

    w, _ = text.size(update_context)
    assert w == 100


def test_text_width_min_height(update_context):
    style = {"min-width": 0, "min-height": 40}
    text = InlineStyle(style, Text("some text"))

    _, h = text.size(update_context)
    assert h == 40


def test_text_with_after_pseudo_element():
    css = """
    ::after { content: " text" }
    """

    class DummyStyleNode:
        name = "text"
        pseudo = None

        def parent(self):
            return None

    style = CompiledStyleSheet(css).compute_style(DummyStyleNode())
    text = Text("some")

    assert text.text(style) == "some text"


def test_iterate_css_node(update_context):
    shape = Box(
        CssNode("first", None, Box(CssNode("nested", None, Text("nested")))),
        CssNode("second", None, Box()),
    )

    assert [
        n.style_node(update_context.style).name() for n in traverse_css_nodes(shape)
    ] == [
        "first",
        "second",
    ]



import io
import textwrap
from xml.sax import SAXParseException

import pytest

import gaphor.storage as storage
from gaphor.storage.parser import MergeConflictDetected, ParserException


def buffer(text):
    file = io.StringIO()
    file.write(textwrap.dedent(text))
    file.seek(0)
    return file


def test_load_model(element_factory, modeling_language):
    file = buffer(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.12.1">
          <StyleSheet id="58d6989a-66f8-11ec-b4c8-0456e5e540ed" />
          <Diagram id="58d6c536-66f8-11ec-b4c8-0456e5e540ed">
            <name>
              <val>main</val>
            </name>
          </Diagram>
        </gaphor>
        """
    )

    storage.load(file, element_factory, modeling_language)

    assert element_factory.lselect()


def test_load_model_with_unknown_element(element_factory, modeling_language):
    file = buffer(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.12.1">
          <StyleSheet id="58d6989a-66f8-11ec-b4c8-0456e5e540ed" />
          <FooBar id="58d6989a-66f8-11ec-b4c8-0456e5e540ed" />
        </gaphor>
        """
    )

    with pytest.raises(storage.UnknownModelElementError):
        storage.load(file, element_factory, modeling_language)

    assert not element_factory.lselect()


def test_plain_text(element_factory, modeling_language):
    file = buffer(
        """\
        Hello world
        """
    )

    with pytest.raises(SAXParseException):
        storage.load(file, element_factory, modeling_language)

    assert not element_factory.lselect()


def test_wrong_tag(element_factory, modeling_language):
    file = buffer(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <not-gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0">
          <FooBar id="58d6989a-66f8-11ec-b4c8-0456e5e540ed" />
        </not-gaphor>
        """
    )

    with pytest.raises(ParserException):
        storage.load(file, element_factory, modeling_language)

    assert not element_factory.lselect()


def test_xml_not_gaphor(element_factory, modeling_language):
    file = buffer(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <html>
          <p id="58d6989a-66f8-11ec-b4c8-0456e5e540ed">Hello</p>
        </html>
        """
    )

    with pytest.raises(ParserException):
        storage.load(file, element_factory, modeling_language)

    assert not element_factory.lselect()


def test_detect_merge_conflict(element_factory, modeling_language):
    file = buffer(
        """\
        <?xml version="1.0" encoding="utf-8"?>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.12.1">
        <StyleSheet id="58d6989a-66f8-11ec-b4c8-0456e5e540ed" />
        <Diagram id="58d6c536-66f8-11ec-b4c8-0456e5e540ed">
          <name>
        <<<<<<< HEAD
            <val>old</val>
        =======
            <val>new</val>
        >>>>>>> 12345678 (incoming change)
            <val>main</val>
          </name>
        </Diagram>
        </gaphor>
        """
    )

    with pytest.raises(MergeConflictDetected):
        storage.load(file, element_factory, modeling_language)

    assert not element_factory.lselect()



import pytest

from gaphor.core.changeset.apply import get_value_change_property_value
from gaphor.core.changeset.compare import RefChange, UnmatchableModel, compare
from gaphor.core.modeling import (
    ElementFactory,
    PendingChange,
    StyleSheet,
)
from gaphor.diagram.general.simpleitem import Box
from gaphor.UML import Class, Diagram, Element, Property


@pytest.fixture
def current(element_factory):
    return element_factory


@pytest.fixture
def ancestor():
    return ElementFactory()


@pytest.fixture
def incoming():
    return ElementFactory()


def test_empty_element_factories(current, ancestor, incoming):
    change_set = list(compare(current, ancestor, incoming))

    assert not change_set


def test_similar_element_factories(current, ancestor, incoming):
    element = incoming.create(Element)
    ancestor.create_as(Element, element.id)

    change_set = list(compare(current, ancestor, incoming))

    assert not change_set


def test_added_element(current, ancestor, incoming, saver):
    diagram = incoming.create(Diagram)

    change = next(compare(current, ancestor, incoming))

    assert change.op == "add"
    assert change.element_id == diagram.id
    assert change.element_name == "Diagram"

    with open("conflict.gaphor", "w", encoding="utf-8") as f:
        f.write(saver())


def test_added_presentation(current, ancestor, incoming):
    ancestor_diagram = ancestor.create(Diagram)
    diagram = incoming.create_as(Diagram, ancestor_diagram.id)
    box = diagram.create(Box)

    change = next(compare(current, ancestor, incoming))

    assert change.op == "add"
    assert change.element_id == box.id
    assert change.element_name == "Box"
    assert change.diagram_id == diagram.id


def test_added_element_with_attribute(current, ancestor, incoming):
    diagram = incoming.create(Diagram)
    diagram.name = "Foo"

    vals = list(compare(current, ancestor, incoming))
    elem_change, attr_change = vals

    assert elem_change.op == "add"
    assert elem_change.element_id == diagram.id
    assert elem_change.element_name == "Diagram"

    assert attr_change.op == "update"
    assert attr_change.element_id == diagram.id
    assert attr_change.property_name == "name"
    assert attr_change.property_value == "Foo"


def test_added_element_with_reference(current, ancestor, incoming):
    diagram = incoming.create(Diagram)
    element = incoming.create(Element)
    ancestor.create_as(Element, element.id)
    diagram.element = element

    change, ref_change, other_ref = list(compare(current, ancestor, incoming))

    assert change.op == "add"
    assert change.element_id == diagram.id
    assert change.element_name == "Diagram"

    assert ref_change.op == "update"
    assert ref_change.element_id == diagram.id
    assert ref_change.property_name == "element"
    assert ref_change.property_ref == element.id

    assert other_ref.op == "add"
    assert other_ref.element_id == element.id
    assert other_ref.property_name == "ownedDiagram"
    assert other_ref.property_ref == diagram.id


def test_removed_element(current, ancestor, incoming):
    diagram = ancestor.create(Diagram)
    diagram.name = "Foo"

    change = next(compare(current, ancestor, incoming))

    assert change.op == "remove"
    assert change.element_id == diagram.id
    assert change.element_name == "Diagram"


@pytest.mark.parametrize(
    "type,name,current_value,incoming_value,expected_value",
    [
        [Diagram, "name", "Old", "New", "New"],
        [Diagram, "name", None, "New", "New"],
        [Diagram, "name", "Old", None, None],
        [Class, "isAbstract", False, True, True],
        [Class, "isAbstract", True, False, None],
    ],
)
def test_changed_value(
    current,
    ancestor,
    incoming,
    type,
    name,
    current_value,
    incoming_value,
    expected_value,
):
    ancestor_element = ancestor.create(type)
    setattr(ancestor_element, name, current_value)
    incoming_element = incoming.create_as(type, ancestor_element.id)
    setattr(incoming_element, name, incoming_value)

    change = next(compare(current, ancestor, incoming))

    assert change.op == "update"
    assert change.element_id == ancestor_element.id
    assert change.property_name == name
    assert get_value_change_property_value(change) == expected_value


def test_changed_enumeration_with_default_value(current, ancestor, incoming):
    ancestor_property = ancestor.create(Property)
    incoming.create_as(Property, ancestor_property.id)
    ancestor_property.aggregation = "shared"

    change = next(compare(current, ancestor, incoming))

    assert change.op == "update"
    assert change.element_id == ancestor_property.id
    assert change.property_name == "aggregation"
    assert change.property_value is None


def test_add_reference(current, ancestor, incoming):
    ancestor_diagram = ancestor.create(Diagram)
    ancestor_element = ancestor.create(Element)
    incoming_diagram = incoming.create_as(Diagram, ancestor_diagram.id)
    incoming_element = incoming.create_as(Element, ancestor_element.id)
    incoming_diagram.element = incoming_element

    add_ref, update_ref = sorted(
        compare(current, ancestor, incoming), key=lambda c: c.op
    )

    assert isinstance(add_ref, RefChange)
    assert add_ref.op == "add"
    assert add_ref.element_id == incoming_element.id
    assert add_ref.property_name == "ownedDiagram"
    assert add_ref.property_ref == ancestor_diagram.id

    assert isinstance(update_ref, RefChange)
    assert update_ref.op == "update"
    assert update_ref.element_id == ancestor_diagram.id
    assert update_ref.property_name == "element"
    assert update_ref.property_ref == incoming_element.id


def test_remove_reference(current, ancestor, incoming):
    ancestor_diagram = ancestor.create(Diagram)
    ancestor_element = ancestor.create(Element)
    ancestor_diagram.element = ancestor_element
    incoming.create_as(Diagram, ancestor_diagram.id)
    incoming_element = incoming.create_as(Element, ancestor_element.id)

    remove_ref, other_ref = sorted(
        compare(current, ancestor, incoming), key=lambda c: c.op
    )

    assert remove_ref.op == "remove"
    assert remove_ref.element_id == incoming_element.id
    assert remove_ref.property_name == "ownedDiagram"
    assert remove_ref.property_ref == ancestor_diagram.id

    assert other_ref.op == "update"
    assert other_ref.element_id == ancestor_diagram.id
    assert other_ref.property_name == "element"
    assert other_ref.property_ref is None


def test_remove_with_common_reference(current, ancestor, incoming):
    ancestor_diagram = ancestor.create(Diagram)
    ancestor_element = ancestor.create(Element)
    ancestor_diagram.element = ancestor_element
    ancestor_common = ancestor.create(Diagram)
    ancestor_common.element = ancestor_element
    incoming.create_as(Diagram, ancestor_diagram.id)
    incoming_element = incoming.create_as(Element, ancestor_element.id)
    incoming_common = incoming.create_as(Diagram, ancestor_common.id)
    incoming_common.element = incoming_element

    remove_ref, other_ref = sorted(
        compare(current, ancestor, incoming), key=lambda c: c.op
    )

    assert remove_ref.op == "remove"
    assert remove_ref.element_id == incoming_element.id
    assert remove_ref.property_name == "ownedDiagram"
    assert remove_ref.property_ref == ancestor_diagram.id

    assert other_ref.op == "update"
    assert other_ref.element_id == ancestor_diagram.id
    assert other_ref.property_name == "element"
    assert other_ref.property_ref is None


def test_pending_changes_end_up_in_current_model(current, ancestor, incoming):
    diagram = incoming.create(Diagram)
    element = incoming.create(Element)
    ancestor.create_as(Element, element.id)
    diagram.element = element

    changes = set(compare(current, ancestor, incoming))

    assert changes == set(current.select(PendingChange))


def test_types_should_match(current, ancestor, incoming):
    ancestor_diagram = ancestor.create(Diagram)
    incoming.create_as(Element, ancestor_diagram.id)

    with pytest.raises(UnmatchableModel):
        next(compare(current, ancestor, incoming))


def test_style_sheet_with_different_ids(current, ancestor, incoming):
    ancestor.create(StyleSheet)
    incoming.create(StyleSheet)

    changes = set(compare(current, ancestor, incoming))

    assert not changes


def test_style_sheet_comparison(current, ancestor, incoming):
    ancestor_style_sheet = ancestor.create(StyleSheet)
    incoming_style_sheet = incoming.create(StyleSheet)
    incoming_style_sheet.styleSheet = "foo {}"

    change = next(compare(current, ancestor, incoming))

    assert change.op == "update"
    assert change.element_id == ancestor_style_sheet.id
    assert change.property_name == "styleSheet"
    assert change.property_value == "foo {}"



import pytest

from gaphor import UML
from gaphor.C4Model import c4model
from gaphor.core.modeling import ElementFactory
from gaphor.core.modeling.elementdispatcher import ElementDispatcher
from gaphor.services.modelinglanguage import ModelingLanguageService
from gaphor.storage.parser import element
from gaphor.storage.tests.test_storage_uml_2_5_upgrade import loader  # noqa: F401


@pytest.fixture
def modeling_language(event_manager):
    return ModelingLanguageService(event_manager)


@pytest.fixture
def element_factory(event_manager, modeling_language):
    element_factory = ElementFactory(
        event_manager, ElementDispatcher(event_manager, modeling_language)
    )
    yield element_factory
    element_factory.shutdown()


@pytest.mark.parametrize(
    "old_name,new_type",
    [
        ["C4Container", c4model.Container],
        ["C4Person", c4model.Person],
        ["C4Database", c4model.Database],
        ["C4Dependency", c4model.Dependency],
    ],
)
def test_c4_elements(old_name, new_type, loader):  # noqa: F811
    parsed_item = element(id="2", type=old_name)
    elem = loader(parsed_item)[0]

    assert elem.__class__ is new_type


def test_uml_dependency(loader):  # noqa: F811
    parsed_item = element(id="2", type="Dependency")
    elem = loader(parsed_item)[0]

    assert elem.__class__ is UML.Dependency



from gaphas.segment import Segment

from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.segment import PresentationSegment


def test_presentation_segment_is_selected(diagram):
    line = diagram.create(LinePresentation)

    segment = Segment(line, diagram)

    assert isinstance(segment, PresentationSegment)



import asyncio
from pathlib import Path

import pytest

from gaphor.diagram.tests.fixtures import find
from gaphor.UML.general.generaltoolbox import general_tools
from gaphor.UML.general.image import ImageItem
from gaphor.UML.general.imagepropertypage import ImagePropertyPage


@pytest.mark.asyncio
async def test_picture_property_select_opens_dialog(
    monkeypatch, diagram, event_manager
):
    # Init objects
    picture = diagram.create(ImageItem)
    property_page = ImagePropertyPage(picture, event_manager)
    widget = property_page.construct()

    # Prepare mocking
    called = False

    async def call(*args, **kwargs):
        nonlocal called
        called |= True

    monkeypatch.setattr("gaphor.UML.general.imagepropertypage.open_file_dialog", call)

    # Emulate the button click event
    button_widget = find(widget, "select-picture-button")
    button_widget.set_activate_signal_from_name("clicked")

    # Test code
    button_widget.activate()

    await asyncio.gather(property_page.background_task)

    assert called is True


@pytest.mark.asyncio
async def test_picture_property_select_valid_name(monkeypatch, diagram, event_manager):
    # Init objects
    picture = next(
        tl for tl in general_tools.tools if tl.id == "toolbox-image"
    ).item_factory(diagram)
    property_page = ImagePropertyPage(picture, event_manager)

    # Prepare mocking
    called = False

    async def error_dialog(*args, **kwargs):
        nonlocal called
        called |= True

    monkeypatch.setattr(
        "gaphor.UML.general.imagepropertypage.error_dialog", error_dialog
    )

    async def open_file_dialog(*args, **kwargs):
        return Path("data/logos/gaphor-24x24.png")

    monkeypatch.setattr(
        "gaphor.UML.general.imagepropertypage.open_file_dialog", open_file_dialog
    )

    await property_page.open_file(None)

    assert called is False
    assert picture.subject.name == "gaphor-24x24"


@pytest.mark.asyncio
async def test_picture_property_select_keep_name(monkeypatch, diagram, event_manager):
    # Init objects
    picture = next(
        tl for tl in general_tools.tools if tl.id == "toolbox-image"
    ).item_factory(diagram)
    property_page = ImagePropertyPage(picture, event_manager)

    # Prepare mocking
    called = False

    async def call(*args, **kwargs):
        nonlocal called
        called |= True

    monkeypatch.setattr("gaphor.UML.general.imagepropertypage.error_dialog", call)

    async def open_file_dialog(*args, **kwargs):
        return Path("data/logos/gaphor-24x24.png")

    monkeypatch.setattr(
        "gaphor.UML.general.imagepropertypage.open_file_dialog", open_file_dialog
    )

    # Test code
    picture.subject.name = "old_name"
    await property_page.open_file(None)

    assert called is False
    assert picture.subject.name == "old_name"


def test_picture_property_select_replace_name_chars(diagram, event_manager):
    # Init objects
    property_page = ImagePropertyPage(None, event_manager)

    # Test code
    temp_image = Path("gaphor/UML/general/tests/test.+gaphor$48-48.png")
    image_name = property_page.sanitize_image_name(temp_image)

    assert image_name == "test__gaphor_48-48"


def test_picture_property_select_empty_name(diagram, event_manager):
    # Init objects
    property_page = ImagePropertyPage(None, event_manager)

    # Test code
    temp_image = Path("gaphor/UML/general/tests/.png")
    image_name = property_page.sanitize_image_name(temp_image)

    assert image_name == "_png"


def test_picture_property_select_empty_extension(diagram, event_manager):
    # Init objects
    property_page = ImagePropertyPage(None, event_manager)

    # Test code
    temp_image = Path("gaphor/UML/general/tests/test_png")
    image_name = property_page.sanitize_image_name(temp_image)

    assert image_name == "test_png"



"""This script lists classes and optionally attributes from UML model created
with Gaphor."""

# ruff: noqa: T201

import optparse
import sys

from gaphor import UML
from gaphor.application import Session

# Setup command line options.
usage = "usage: %prog [options] file.gaphor"


def main():
    parser = optparse.OptionParser(usage=usage)

    parser.add_option(
        "-a",
        "--attributes",
        dest="attrs",
        action="store_true",
        help="Print class attributes",
    )

    (options, args) = parser.parse_args()

    if len(args) != 1:
        parser.print_help()
        sys.exit(1)

    # The model file to load.
    model = args[0]

    # Create the Gaphor application object.
    session = Session()

    # Get services we need.
    element_factory = session.get_service("element_factory")
    file_manager = session.get_service("file_manager")

    # Load model from file.
    file_manager.load(model)

    # Find all classes using factory select.
    for cls in element_factory.select(UML.Class):
        print(f"Found class {cls.name}")

        if options.attrs:
            for attr in cls.ownedAttribute:
                print(f" Attribute: {attr.name}")


if __name__ == "__main__":
    main()



from gaphor import UML
from gaphor.diagram.tests.fixtures import connect, copy_clear_and_paste_link
from gaphor.UML.states import StateItem, TransitionItem


def test_state_with_behaviors(diagram, element_factory):
    state: UML.State = element_factory.create(UML.State)
    state.entry = element_factory.create(UML.Behavior)
    state.entry.name = "hi"
    state.exit = element_factory.create(UML.Behavior)
    state.exit.name = "bye"
    state.doActivity = element_factory.create(UML.Behavior)
    state.doActivity.name = "act"

    state_item = diagram.create(StateItem, subject=state)

    new_items = copy_clear_and_paste_link({state_item}, diagram, element_factory)
    new_state_item = new_items.pop()

    assert isinstance(new_state_item, StateItem)
    assert new_state_item.subject.entry.name == "hi"
    assert new_state_item.subject.exit.name == "bye"
    assert new_state_item.subject.doActivity.name == "act"


def test_connected_transition(diagram, element_factory):
    state_item_1 = diagram.create(StateItem, subject=element_factory.create(UML.State))
    state_item_2 = diagram.create(StateItem, subject=element_factory.create(UML.State))
    transition_item = diagram.create(TransitionItem)

    connect(transition_item, transition_item.head, state_item_1)
    connect(transition_item, transition_item.tail, state_item_2)
    specification = element_factory.create(UML.LiteralString)
    specification.value = "[test]"
    specification.owningConstraint = transition_item.subject.guard
    transition_item.subject.guard.specification = specification

    new_items = copy_clear_and_paste_link({transition_item}, diagram, element_factory)
    new_transition_item = new_items.pop()

    assert isinstance(new_transition_item, TransitionItem), new_items
    assert new_transition_item.subject.guard
    assert new_transition_item.subject.guard.specification.value == "[test]"



import pytest

from gaphor.application import Application
from gaphor.core import event_handler
from gaphor.event import ActiveSessionChanged, ApplicationShutdown, SessionShutdown


@pytest.fixture
def application():
    application = Application()
    yield application
    application.shutdown()


def two_sessions(application):
    session1 = application.new_session(
        services=["component_registry", "event_manager", "tools_menu"]
    )

    session2 = application.new_session(
        services=["component_registry", "event_manager", "tools_menu"]
    )

    return session1, session2


def test_most_recently_created_session_is_active(application):
    _session1, session2 = two_sessions(application)

    assert application.active_session is session2


def test_active_window_changed(application):
    session1, _session2 = two_sessions(application)

    session1.get_service("event_manager").handle(ActiveSessionChanged(None))

    assert application.active_session is session1


def test_session_shutdown(application):
    session1, session2 = two_sessions(application)

    session2.get_service("event_manager").handle(SessionShutdown(quitting=False))

    assert len(application.sessions) == 1
    assert session1 in application.sessions


def test_all_sessions_shut_down(application):
    quit_events = []

    @event_handler(ApplicationShutdown)
    def on_quit(event):
        quit_events.append(event)

    application.event_manager.subscribe(on_quit)

    session1, session2 = two_sessions(application)

    session1.get_service("event_manager").handle(SessionShutdown(quitting=False))
    session2.get_service("event_manager").handle(SessionShutdown(quitting=False))

    assert len(application.sessions) == 0
    assert quit_events



from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.storage.load import load
from gaphor.UML import diagramitems


def test_message_item_upgrade(element_factory, modeling_language, test_models):
    with (test_models / "multiple-messages.gaphor").open(encoding="utf-8") as f:
        load(f, element_factory, modeling_language)

    diagram = element_factory.lselect(Diagram)[0]
    items = [e for e in diagram.get_all_items() if not e.parent]
    message_items = [i for i in items if isinstance(i, diagramitems.MessageItem)]
    subjects = [m.subject for m in message_items]
    messages = element_factory.lselect(UML.Message)
    presentations = [m.presentation for m in messages]

    assert len(messages) == 10
    assert all(subjects), subjects
    assert len(message_items) == 10
    assert all(presentations), presentations



from gaphor import UML
from gaphor.UML.classes.enumerationpropertypages import (
    EnumerationPage,
    ShowEnumerationPage,
)


def test_enumeration_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.Enumeration)

    property_page = EnumerationPage(subject, event_manager)

    property_page.construct()


def test_show_enumeration_property_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.classes.EnumerationItem, subject=element_factory.create(UML.Enumeration)
    )
    property_page = ShowEnumerationPage(item, event_manager)

    property_page.construct()



"""Test if models are up to date."""

import functools
from pathlib import Path

import pytest

from gaphor.C4Model import c4model
from gaphor.codegen import coder
from gaphor.codegen.coder import main
from gaphor.core.modeling import coremodel
from gaphor.RAAML import raaml
from gaphor.SysML import sysml
from gaphor.UML import uml


@pytest.fixture(scope="session", autouse=True)
def cache_loaded_models():
    """Speed up testing by caching loaded models."""
    coder.load_model = functools.lru_cache(maxsize=None)(coder.load_model)


def test_core_model(tmp_path):
    outfile = tmp_path / "coremodel.py"
    main(
        modelfile="models/Core.gaphor",
        overridesfile="models/Core.override",
        outfile=outfile,
    )

    current_model = Path(coremodel.__file__).read_text(encoding="utf-8")
    generated_model = outfile.read_text(encoding="utf-8")

    assert generated_model == current_model


def test_uml_model(tmp_path):
    outfile = tmp_path / "uml.py"
    main(
        modelfile="models/UML.gaphor",
        overridesfile="models/UML.override",
        supermodelfiles=[("Core", "models/Core.gaphor")],
        outfile=outfile,
    )

    current_model = Path(uml.__file__).read_text(encoding="utf-8")
    generated_model = outfile.read_text(encoding="utf-8")

    assert generated_model == current_model


def test_c4model_model(tmp_path):
    outfile = tmp_path / "c4model.py"
    main(
        modelfile="models/C4Model.gaphor",
        supermodelfiles=[
            ("UML", "models/UML.gaphor"),
            ("Core", "models/Core.gaphor"),
        ],
        outfile=outfile,
    )

    current_model = Path(c4model.__file__).read_text(encoding="utf-8")
    generated_model = outfile.read_text(encoding="utf-8")

    assert generated_model == current_model


def test_sysml_model(tmp_path):
    outfile = tmp_path / "sysml.py"
    main(
        modelfile="models/SysML.gaphor",
        overridesfile="models/SysML.override",
        supermodelfiles=[
            ("UML", "models/UML.gaphor"),
            ("Core", "models/Core.gaphor"),
        ],
        outfile=outfile,
    )

    current_model = Path(sysml.__file__).read_text(encoding="utf-8")
    generated_model = outfile.read_text(encoding="utf-8")

    assert generated_model == current_model


def test_raaml_model(tmp_path):
    outfile = tmp_path / "raaml.py"
    main(
        modelfile="models/RAAML.gaphor",
        supermodelfiles=[
            ("SysML", "models/SysML.gaphor"),
            ("UML", "models/UML.gaphor"),
            ("Core", "models/Core.gaphor"),
        ],
        outfile=outfile,
    )

    current_model = Path(raaml.__file__).read_text(encoding="utf-8")
    generated_model = outfile.read_text(encoding="utf-8")

    assert generated_model == current_model



from gaphor import C4Model
from gaphor.C4Model.propertypages import DescriptionPropertyPage, TechnologyPropertyPage
from gaphor.diagram.tests.fixtures import find


def test_description_property_page(event_manager, element_factory):
    subject = element_factory.create(C4Model.c4model.Container)
    property_page = DescriptionPropertyPage(subject, event_manager)

    widget = property_page.construct()
    description = find(widget, "description")
    description.get_buffer().set_text("test")

    assert subject.description == "test"


def test_technology_property_page(event_manager, element_factory):
    subject = element_factory.create(C4Model.c4model.Container)
    property_page = TechnologyPropertyPage(subject, event_manager)

    widget = property_page.construct()
    technology = find(widget, "technology")
    technology.set_text("test")

    assert subject.technology == "test"



from gaphor.core.modeling.base import Base

__modeling_language__ = "test"


class A(Base):
    pass


def test_modeling_language():
    assert Base.__modeling_language__ == "Core"
    assert A.__modeling_language__ == "test"



# ruff: noqa: F401,F811
import gaphor.storage as storage
from gaphor import UML
from gaphor.conftest import element_factory, event_manager, modeling_language
from gaphor.core.modeling import Diagram
from gaphor.UML.actions import ActionItem, ControlFlowItem


def test_action_issue(element_factory, modeling_language, test_models):
    """Test an issue when loading a freshly created action diagram."""
    with (test_models / "action-issue.gaphor").open(encoding="utf-8") as f:
        storage.load(f, element_factory, modeling_language)

    actions = element_factory.lselect(UML.Action)
    flows = element_factory.lselect(UML.ControlFlow)
    assert 3 == len(actions)
    assert 3 == len(flows)

    # Actions live in partitions:
    partitions = element_factory.lselect(UML.ActivityPartition)
    assert 2 == len(partitions)

    # Okay, so far the data model is saved correctly. Now, how do the
    # handles behave?
    diagrams = element_factory.lselect(Diagram)
    assert 1 == len(diagrams)

    diagram = diagrams[0]
    assert 7 == len(list(diagram.get_all_items()))
    # Part, Act, Act, Act, Flow, Flow, Flow

    for e in actions + flows:
        assert 1 == len(e.presentation), e
    for i in diagram.select(lambda e: isinstance(e, ControlFlowItem | ActionItem)):
        assert i.subject, i

    # Loaded as:
    #
    # actions[0] --> flows[0, 1]
    # flows[0, 2] --> actions[1]
    # flows[1] --> actions[2] --> flows[2]

    # start element:
    assert actions[0].outgoing[0] is flows[0]
    assert actions[0].outgoing[1] is flows[1]
    assert actions[2].outgoing[0] is flows[2]
    assert not actions[0].incoming

    (cinfo,) = diagram.connections.get_connections(handle=flows[0].presentation[0].head)
    assert cinfo.connected is actions[0].presentation[0]
    (cinfo,) = diagram.connections.get_connections(handle=flows[1].presentation[0].head)
    assert cinfo.connected is actions[0].presentation[0]

    # Intermediate element:
    assert actions[2].incoming[0] is flows[1]
    assert actions[2].outgoing[0] is flows[2]

    (cinfo,) = diagram.connections.get_connections(handle=flows[1].presentation[0].tail)
    assert cinfo.connected is actions[2].presentation[0]
    (cinfo,) = diagram.connections.get_connections(handle=flows[2].presentation[0].head)
    assert cinfo.connected is actions[2].presentation[0]

    # Final element:
    assert actions[1].incoming[0] is flows[0]
    assert actions[1].incoming[1] is flows[2]

    (cinfo,) = diagram.connections.get_connections(handle=flows[0].presentation[0].tail)
    assert cinfo.connected is actions[1].presentation[0]
    (cinfo,) = diagram.connections.get_connections(handle=flows[2].presentation[0].tail)
    assert cinfo.connected is actions[1].presentation[0]

    # Test the parent-child connectivity
    for a in actions:
        (p,) = a.inPartition
        assert p
        assert a.presentation[0].parent
        assert a.presentation[0].parent is p.presentation[0]



from __future__ import annotations

import enum

import pytest

from gaphor.core import event_handler
from gaphor.core.modeling import Base
from gaphor.core.modeling.collection import collection
from gaphor.core.modeling.event import AssociationUpdated
from gaphor.core.modeling.properties import (
    UnlimitedNatural,
    association,
    attribute,
    derived,
    derivedunion,
    enumeration,
    relation_many,
    relation_one,
)


def test_association_1_x():
    #
    # 1:-
    #
    class A(Base):
        one: relation_one[B | None]

    class B(Base):
        two: relation_one[A]

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1)
    a = A()
    b = B()
    a.one = b
    assert a.one is b
    assert b.two is a

    a.one = None
    assert a.one is None
    assert b.two is None

    a.one = b
    assert a.one is b
    assert b.two is a

    del a.one
    assert a.one is None
    assert b.two is None


def test_association_n_1():
    class A(Base):
        one: relation_many[B]

    class B(Base):
        two: relation_one[A]

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, 1)

    a = A()
    b = B()
    a.one = b
    assert b in a.one
    assert b.two is a


def test_reassign_association_n_1():
    class A(Base):
        many: relation_many[B]

    class B(Base):
        one: relation_one[A]

    A.many = association("many", B, 0, "*", opposite="one")
    B.one = association("one", A, 0, 1, opposite="many")
    a = A()
    aa = A()
    b = B()
    a.many = b
    aa.many = b
    assert b in aa.many
    assert not a.many
    assert b.one is aa


def test_association_load_does_not_steal_references():
    class A(Base):
        many: relation_many[B]

    class B(Base):
        one: relation_one[A]

    A.many = association("many", B, 0, "*", opposite="one")
    B.one = association("one", A, 0, 1, opposite="many")
    a = A()
    aa = A()
    b = B()
    a.many = b

    A.many.load(aa, b)

    assert b in a.many
    assert not aa.many
    assert b.one is a


def test_association_1_1():
    class A(Base):
        one: relation_one[B]

    class B(Base):
        two: relation_one[A]

    class C(Base):
        pass

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1, opposite="one")

    a = A()
    b = B()
    a.one = b
    a.one = b

    assert a.one is b
    assert b.two is a

    a.one = B()
    assert a.one is not b
    assert b.two is None

    c = C()

    with pytest.raises(TypeError):
        a.one = c

    assert a.one is not c

    del a.one
    assert a.one is None
    assert b.two is None


def test_association_1_n():
    #
    # 1:n
    #
    class A(Base):
        one: relation_one[B]

    class B(Base):
        two: relation_many[A]

    class C(Base):
        pass

    A.one = association("one", B, lower=0, upper=1, opposite="two")
    B.two = association("two", A, lower=0, upper="*", opposite="one")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    b1.two = a1
    assert len(b1.two) == 1, f"len(b1.two) == {len(b1.two)}"
    assert a1 in b1.two
    assert a1.one is b1, f"{a1.one}/{b1}"
    b1.two = a1
    b1.two = a1
    assert len(b1.two) == 1, f"len(b1.two) == {len(b1.two)}"
    assert a1 in b1.two
    assert a1.one is b1, f"{a1.one}/{b1}"

    b1.two = a2
    assert a1 in b1.two
    assert a2 in b1.two
    assert a1.one is b1
    assert a2.one is b1

    try:
        del b1.two
    except Exception:
        pass  # ok
    else:
        assert b1.two != []

    assert a1 in b1.two
    assert a2 in b1.two
    assert a1.one is b1
    assert a2.one is b1

    b1.two.remove(a1)

    assert len(b1.two) == 1
    assert a1 not in b1.two
    assert a2 in b1.two
    assert a1.one is None
    assert a2.one is b1

    a2.one = b2

    assert len(b1.two) == 0
    assert len(b2.two) == 1
    assert a2 in b2.two
    assert a1.one is None
    assert a2.one is b2

    del b1.two[a1]


def test_association_n_n():
    """Test association n:n."""

    class A(Base):
        one: relation_many[B]

    class B(Base):
        two: relation_many[A]

    class C(Base):
        pass

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, "*", opposite="one")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    assert b1 in a1.one
    assert a1 in b1.two
    assert not a2.one
    assert not b2.two

    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two
    assert not a2.one

    a2.one = b1
    assert len(a1.one) == 2
    assert len(a2.one) == 1
    assert len(b1.two) == 2
    assert len(b2.two) == 1
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two
    assert b1 in a2.one
    assert a2 in b1.two

    del a1.one[b1]
    assert len(a1.one) == 1
    assert len(a2.one) == 1
    assert len(b1.two) == 1
    assert len(b2.two) == 1
    assert b1 not in a1.one
    assert b2 in a1.one
    assert a1 not in b1.two
    assert a1 in b2.two
    assert b1 in a2.one
    assert a2 in b1.two


def test_association_swap():
    class A(Base):
        one: relation_many[B]

    class B(Base):
        pass

    class C(Base):
        pass

    A.one = association("one", B, 0, "*")

    a = A()
    b1 = B()
    b2 = B()

    a.one = b1
    a.one = b2
    assert a.one.size() == 2
    assert a.one[0] is b1
    assert a.one[1] is b2

    events: list[object] = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    a.one.swap(b1, b2)

    assert a.one.size() == 2
    assert a.one[0] is b2
    assert a.one[1] is b1


def test_association_unlink_1():
    class A(Base):
        one: relation_many[B]

    class B(Base):
        pass

    class C(Base):
        pass

    A.one = association("one", B, 0, "*")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one

    a2.one = b1

    # remove b1 from all elements connected to b1
    # also the signal should be removed
    b1.unlink()

    assert b1 not in a1.one
    assert b2 in a1.one


def test_association_unlink_2():
    #
    # unlink
    #
    class A(Base):
        one: relation_many[B]

    class B(Base):
        two: relation_many[A]

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, "*")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two

    a2.one = b1

    # remove b1 from all elements connected to b1
    # also the signal should be removed
    b1.unlink()

    assert b1 not in a1.one
    assert b2 in a1.one
    assert a1 not in b1.two
    assert a1 in b2.two


def test_association_subsettable_0_n(element_factory):
    class A(Base):
        pass

    class C(Base):
        original: relation_many[A]
        subset: relation_many[A]

    C.original = association("original", A, 0, "*")
    C.subset = association("subset", A, 0, "*")

    C.original.add(C.subset)

    a1 = A()
    a2 = A()

    # Subset changes, superset has multiplicity *, subset has multiplicity 1 or *:
    #     Both sets {}: subset {a1} => superset {a1}
    c1 = C()
    assert len(c1.original) == 0
    assert len(c1.subset) == 0
    c1.subset = a1
    assert len(c1.original) == 1
    assert a1 in c1.original

    #     Superset {a1}, subset {}, subset {a2} => superset {a1, a2}
    c2 = C()
    c2.original = a1
    assert len(c2.subset) == 0
    c2.subset = a2
    assert len(c2.original) == 2
    assert a1 in c2.original
    assert a2 in c2.original

    #     Superset {a1}, subset {a1}: subset {a2}, => superset {a1, a2} ***POLICY***
    c3 = C()
    c3.subset = a1
    assert a1 in c3.original
    c3.subset.remove(a1)
    c3.subset = a2
    assert len(c3.original) == 2
    assert a2 in c3.original

    #     Superset {a1}, subset {a1}: subset {} => superset {a1} ***POLICY***
    c4 = C()
    c4.original = a1
    c4.subset = a1
    c4.subset.remove(a1)
    assert len(c4.original) == 1

    # Superset changes, superset has multiplicity *, subset has multiplicity 1 or *:
    #     Both sets {}: superset {a1} => subset {}
    c5 = C()
    c5.original = a1
    assert len(c5.subset) == 0

    #     Subset {a1}, superset {a1}: superset {} => subset {}
    c6 = C()
    c6.subset = a1
    assert len(c6.original) == 1
    c6.original.remove(a1)
    assert len(c6.subset) == 0

    #     Superset {a1}, subset {a1}, superset {a2} => subset {} ***POLICY***
    c7 = C()
    c7.subset = a1
    assert a1 in c7.original
    c7.original.remove(a1)
    c7.original = a2
    assert len(c7.subset) == 0


def test_association_subsettable_0_1(element_factory):
    class A(Base):
        pass

    class C(Base):
        original: relation_many[A]
        subset: relation_many[A]

    C.original = association("original", A, 0, 1)
    C.subset = association("subset", A, 0, 1)

    C.original.add(C.subset)

    a1 = A()
    a2 = A()

    # Subset changes, superset has multiplicity 0..1, subset has multiplicity 0..1:
    #     Both sets {}: subset {a1} => superset {a1}
    c1 = C()
    assert c1.original is None
    assert c1.subset is None
    c1.subset = a1
    assert c1.original is a1

    #     Superset {a1}, subset {}, subset {a2} => superset {a2}
    c2 = C()
    c2.original = a1
    assert c2.subset is None
    c2.subset = a2
    assert c2.original is a2

    #     Superset {a1}, subset {a1}: subset {a2}, => superset {a2} ***POLICY***
    c3 = C()
    c3.subset = a1
    assert c3.original is a1
    c3.subset = a2
    assert c3.original is a2

    #     Superset {a1}, subset {a1}: subset {} => superset {} ***POLICY***
    c4 = C()
    c4.original = a1
    c4.subset = a1
    c4.subset = None
    assert c4.original is None

    # Superset changes, superset has multiplicity *, subset has multiplicity 1 or *:
    #     Both sets {}: superset {a1} => subset {}
    c5 = C()
    c5.original = a1
    assert c5.subset is None

    #     Subset {a1}, superset {a1}: superset {} => subset {}
    c6 = C()
    c6.subset = a1
    assert c6.original is a1
    c6.original = None
    assert c6.subset is None

    #     Superset {a1}, subset {a1}, superset {a2} => subset {} ***POLICY***
    c7 = C()
    c7.subset = a1
    assert c7.original is a1
    c7.original = a2
    assert c7.subset is None


def test_association_subsettable_updates_derived_union(element_factory):
    class A(Base):
        pass

    class C(Base):
        original: relation_many[A]
        subset: relation_many[A]
        union: relation_many[A]

    C.original = association("original", A, 0, 1)
    C.subset = association("subset", A, 0, 1)
    C.original.add(C.subset)
    C.union = derivedunion("u", object, 0, "*", C.original)

    a1 = A()
    a2 = A()

    c1 = C()
    c1.subset = a1
    assert a1 in c1.union
    c1.subset = a2
    assert a2 in c1.union
    assert a1 not in c1.union


# def test_association_subsettable_add_fails_when_subset_multiplicity_exceeds_superset_multiplicity(element_factory):
#     class A(Base):
#         pass

#     class C(Base):
#         original: relation_many[A]
#         subset: relation_many[A]

#     C.original = association("original", A, 0, 1)
#     C.subset = association("subset", A, 0, "*")

#     with pytest.raises(Exception) as execinfo:
#         C.original.add(C.subset)
#     # assert str(execinfo.value) == "Cannot add a multiplicity * subset to a multiplicity 1 superset"


def test_can_not_set_association_to_owner(element_factory, event_manager):
    class A(Base):
        pass

    A.a = association("a", A, upper=1)

    a = element_factory.create(A)

    with pytest.raises(TypeError):
        a.a = a


def test_attributes():
    class A(Base):
        a: attribute[str]

    A.a = attribute("a", str, "default")

    a = A()
    assert a.a == "default", a.a
    a.a = "bar"
    assert a.a == "bar", a.a
    del a.a
    assert a.a == "default"

    with pytest.raises(TypeError):
        a.a = 1


@pytest.mark.parametrize(
    "input,expected",
    [
        [None, None],
        [0, 0],
        [1, 1],
        [2, 2],
        [False, 0],
        [True, 1],
        ["False", 0],
        ["True", 1],
    ],
)
def test_int_and_boolean_attributes(input, expected):
    class A(Base):
        a = attribute("a", int, 0)

    a = A()
    a.a = input

    assert a.a == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        [None, None],
        [0, 0],
        [1, 1],
        [2, 2],
        ["*", "*"],
    ],
)
def test_unlimited_natural_attributes(input, expected):
    class A(Base):
        a = attribute("a", UnlimitedNatural, 0)

    a = A()
    a.a = input

    assert a.a == expected


@pytest.mark.parametrize(
    "input",
    [
        "foo",
        -1,
    ],
)
def test_unlimited_natural_attributes_wrong_inputs(input):
    class A(Base):
        a = attribute("a", UnlimitedNatural, 0)

    a = A()

    with pytest.raises(ValueError):
        a.a = input


def test_attributes_loading_failure():
    class A(Base):
        a: attribute[int]

    A.a = attribute("a", int, 0)

    a = A()

    with pytest.raises(ValueError):
        A.a.load(a, "not-int")


def test_enumerations():
    class EnumKind(enum.StrEnum):
        one = "one"
        two = "two"
        three = "three"

    class A(Base):
        a: enumeration

    A.a = enumeration("a", EnumKind, EnumKind.one)
    a = A()
    assert a.a == "one"
    a.a = "two"
    assert a.a == "two"
    a.a = "three"
    assert a.a == "three"

    with pytest.raises(TypeError):
        a.a = "four"

    assert a.a == EnumKind.three

    del a.a
    assert a.a == "one"


def test_derived():
    class A(Base):
        a: relation_many[A]

    A.a = derived("a", str, 0, "*", lambda self: ["a", "b", "c"])

    a = A()

    assert isinstance(a.a, collection)
    assert a.a == ["a", "b", "c"]


def test_derived_single_properties():
    class A(Base):
        pass

    class E(Base):
        notified = attribute("notified", int)
        a: relation_one[A]
        u: relation_one[A]

        def handle(self, event):
            if event.property is E.u:
                self.notified = True

    E.a = association("a", A, upper=1)
    E.u = derived("u", A, 0, 1, lambda self: [self.a], E.a)

    e = E()
    e.a = A()

    assert e.a is e.u
    assert e.notified


def test_derived_multi_properties():
    class A(Base):
        pass

    class E(Base):
        notified = attribute("notified", int)
        a: relation_one[A]
        b: relation_many[A]
        u: relation_many[A]

        def handle(self, event):
            if event.property is E.u:
                self.notified = True

    E.a = association("a", A, upper=1)
    E.b = association("b", A)
    E.u = derived("u", A, 0, "*", lambda self: {self.a, *self.b}, E.a, E.b)

    e = E()
    e.a = A()
    e.b = A()

    assert e.a in e.u
    assert e.b[0] in e.u
    assert e.notified


def test_derivedunion():
    class A(Base):
        a: relation_many[A]
        b: relation_one[A]
        u: relation_many[A]

    A.a = association("a", A)
    A.b = association("b", A, 0, 1)
    A.u = derivedunion("u", object, 0, "*", A.a, A.b)

    a = A()
    assert len(a.a) == 0, f"a.a = {a.a}"
    assert len(a.u) == 0, f"a.u = {a.u}"
    a.a = b = A()
    a.a = c = A()
    assert len(a.a) == 2, f"a.a = {a.a}"
    assert b in a.a
    assert c in a.a
    assert len(a.u) == 2, f"a.u = {a.u}"
    assert b in a.u
    assert c in a.u

    a.b = d = A()
    assert len(a.a) == 2, f"a.a = {a.a}"
    assert b in a.a
    assert c in a.a
    assert d == a.b
    assert len(a.u) == 3, f"a.u = {a.u}"
    assert b in a.u
    assert c in a.u
    assert d in a.u


def test_derivedunion_notify_for_single_derived_property():
    class A(Base):
        pass

    class E(Base):
        notified = False

        a: relation_many[A]
        u: relation_many[A]

        def handle(self, event):
            if event.property is E.u:
                self.notified = True

    E.a = association("a", A)
    E.u = derivedunion("u", A, 0, "*", E.a)

    e = E()
    e.a = A()

    assert e.notified is True


def test_derivedunion_notify_for_multiple_derived_properties():
    class A(Base):
        pass

    class E(Base):
        notified = False

        a: relation_many[A]
        aa: relation_many[A]
        u: relation_many[A]

        def handle(self, event):
            if event.property is E.u:
                self.notified = True

    E.a = association("a", A)
    E.aa = association("aa", A)
    E.u = derivedunion("u", A, 0, "*", E.a, E.aa)

    e = E()
    e.a = A()

    assert e.notified is True


def test_derivedunion_notify_for_single_and_multi_derived_properties():
    class A(Base):
        pass

    class E(Base):
        a: relation_one[A]
        b: relation_many[A]
        u: relation_many[A]

    E.a = association("a", A, upper=1)
    E.b = association("b", A)
    E.u = derivedunion("u", A, 0, "*", E.a, E.b)

    e = E()
    e.a = A()
    e.b = A()

    assert e.a in e.u
    assert e.b[0] in e.u


def test_derivedunion_listmixins():
    class A(Base):
        a: relation_many[A]
        b: relation_many[A]
        u: relation_many[A]
        name: attribute[str]

    A.a = association("a", A)
    A.b = association("b", A)
    A.u = derivedunion("u", A, 0, "*", A.a, A.b)
    A.name = attribute("name", str, "default")

    a = A()
    a.a = A()
    a.a = A()
    a.b = A()
    a.a[0].name = "foo"
    a.a[1].name = "bar"
    a.b[0].name = "baz"

    assert list(a.a[:].name) == ["foo", "bar"]
    assert sorted(a.u[:].name) == ["bar", "baz", "foo"]


def test_composite():
    class A(Base):
        is_unlinked = False
        name = attribute("name", str)
        comp: relation_many[A]
        other: relation_many[A]

        def unlink(self):
            self.is_unlinked = True
            Base.unlink(self)

    A.comp = association("comp", A, composite=True, opposite="other")
    A.other = association("other", A, composite=False, opposite="comp")

    a = A()
    a.name = "a"
    b = A()
    b.name = "b"
    a.comp = b
    assert b in a.comp
    assert a in b.other

    a.unlink()
    assert a.is_unlinked
    assert b.is_unlinked



"""Test connections to folded interface."""

from gaphor import UML
from gaphor.core.modeling import StyleSheet
from gaphor.core.modeling.diagram import StyledItem
from gaphor.diagram.tests.fixtures import connect, disconnect
from gaphor.UML.classes.dependency import DependencyItem
from gaphor.UML.classes.interface import Folded, InterfaceItem
from gaphor.UML.classes.interfacerealization import InterfaceRealizationItem
from gaphor.UML.classes.klass import ClassItem


def test_interface_realization_default_line_style(element_factory, diagram):
    style_sheet = element_factory.create(StyleSheet)
    impl = diagram.create(InterfaceRealizationItem)
    style = style_sheet.compute_style(StyledItem(impl))

    assert style["dash-style"]


def test_interface_realization_folded_interface_connection(element_factory, diagram):
    """Test connecting implementation to folded interface."""
    style_sheet = element_factory.create(StyleSheet)
    iface = diagram.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    iface.folded = Folded.PROVIDED
    impl = diagram.create(InterfaceRealizationItem)

    connect(impl, impl.head, iface)
    diagram.update({iface, impl})

    style = style_sheet.compute_style(StyledItem(impl))
    assert not style["dash-style"]


def test_interface_realization_folded_interface_disconnection(element_factory, diagram):
    """Test disconnection implementation from folded interface."""
    style_sheet = element_factory.create(StyleSheet)
    iface = diagram.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    iface.folded = Folded.PROVIDED
    impl = diagram.create(InterfaceRealizationItem)

    connect(impl, impl.head, iface)
    disconnect(impl, impl.head)
    impl.request_update()
    style = style_sheet.compute_style(StyledItem(impl))

    assert style["dash-style"]


def test_dependency_default_line_style(element_factory, diagram):
    style_sheet = element_factory.create(StyleSheet)
    dep = diagram.create(DependencyItem)
    style = style_sheet.compute_style(StyledItem(dep))

    assert style["dash-style"]


def test_dependency_folded_interface_connection(element_factory, diagram):
    """Test connecting dependency to folded interface."""
    style_sheet = element_factory.create(StyleSheet)
    clazz = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    iface = diagram.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    iface.folded = Folded.PROVIDED
    dep = diagram.create(DependencyItem)

    connect(dep, dep.head, iface)
    connect(dep, dep.tail, clazz)
    iface.update_shapes()
    diagram.update({clazz, iface, dep})
    style = style_sheet.compute_style(StyledItem(dep))

    assert dep.subject
    assert not style["dash-style"]
    assert iface.folded == Folded.REQUIRED


def test_dependency_folded_interface_disconnection(element_factory, diagram):
    """Test disconnection dependency from folded interface."""
    style_sheet = element_factory.create(StyleSheet)
    iface = diagram.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    iface.folded = Folded.PROVIDED
    dep = diagram.create(DependencyItem)

    connect(dep, dep.head, iface)
    disconnect(dep, dep.head)
    dep.request_update()
    style = style_sheet.compute_style(StyledItem(dep))

    assert style["dash-style"]
    assert iface.folded == Folded.PROVIDED


def test_dependency_unfolded_interface_connection(element_factory, diagram):
    """Test disconnection dependency from unfolded interface."""
    style_sheet = element_factory.create(StyleSheet)
    iface = diagram.create(InterfaceItem, subject=element_factory.create(UML.Interface))
    dep = diagram.create(DependencyItem)

    connect(dep, dep.head, iface)
    style = style_sheet.compute_style(StyledItem(dep))

    assert (7.0, 5.0) == style["dash-style"]
